from datetime import timedelta

import pytest
from expects import *

from sym.cli.errors import AccessDenied, SuppressedError
from sym.cli.helpers.config import Config
from sym.cli.helpers.ssh import maybe_gen_ssh_key, ssh_key_and_config
from sym.cli.saml_clients.saml_client import SAMLClient
from sym.cli.tests.helpers.capture import CaptureCommand
from sym.cli.tests.helpers.sandbox import Sandbox

INSTANCE_ID = "123"
IP_ADDRESS = "127.0.0.1"


def describe_instances_stub(make_stub):
    ec2 = make_stub("ec2")
    ec2.add_response(
        "describe_instances",
        {"Reservations": [{"Instances": [{"InstanceId": INSTANCE_ID}]}]},
    )


def send_command_stub(make_stub, command_id):
    ssm = make_stub("ssm")
    ssm.add_response(
        "send_command",
        {"Command": {"CommandId": command_id, "Status": "Pending"}},
    )
    ssm.add_response(
        "get_command_invocation",
        {"CommandId": command_id, "Status": "Success"},
        {"CommandId": command_id, "InstanceId": INSTANCE_ID},
    )


@pytest.fixture
def ssh_tester(
    command_tester,
    saml_client: SAMLClient,
    capture_command: CaptureCommand,
    sandbox: Sandbox,
):
    def setup():
        sandbox.link_binary("ssh-keygen")
        capture_command.allow_call("ssh-keygen")

    def teardown():
        for conf in ssh_key_and_config(saml_client):
            expect(conf.path.exists()).to(be_true)

    return command_tester(["ssh", "test", IP_ADDRESS], setup=setup, teardown=teardown)


def test_ssh(ssh_tester, capture_command: CaptureCommand):
    def setup(make_stub):
        describe_instances_stub(make_stub)

    with ssh_tester(setup=setup):
        capture_command.assert_command(
            ["exec", "true"],
            ["ssh-keygen"],
            ["ssh", INSTANCE_ID, "-o BatchMode=yes"],
            ["ssh", INSTANCE_ID],
        )


def test_ssh_with_key(
    ssh_tester, saml_client: SAMLClient, capture_command: CaptureCommand
):
    def setup(make_stub):
        describe_instances_stub(make_stub)
        maybe_gen_ssh_key(saml_client)

    with ssh_tester(setup=setup):
        capture_command.assert_command(
            ["exec", "true"],
            ["ssh", INSTANCE_ID, "-o BatchMode=yes"],
            ["ssh", INSTANCE_ID],
        )


def test_ssh_needs_put(uuid, ssh_tester, capture_command: CaptureCommand):
    def setup(make_stub):
        describe_instances_stub(make_stub)
        send_command_stub(make_stub, str(uuid))

        capture_command.register_output(
            r"BatchMode",
            "Permission denied (publickey)",
            exit_code=255,
        )

    with ssh_tester(setup=setup):
        capture_command.assert_command(
            ["exec", "true"],
            ["ssh-keygen"],
            ["ssh", INSTANCE_ID, "-o BatchMode=yes"],
            ["ssh", INSTANCE_ID],
        )


def test_ssh_unknown_error(ssh_tester, capture_command: CaptureCommand):
    def setup(make_stub):
        describe_instances_stub(make_stub)

        capture_command.register_output(
            r"BatchMode",
            "Random Error",
            exit_code=255,
        )

    with ssh_tester(setup=setup, exception=SuppressedError) as result:
        capture_command.assert_command(
            ["exec", "true"],
            ["ssh-keygen"],
            ["ssh", INSTANCE_ID, "-o BatchMode=yes"],
        )

        expect(result.output).to(equal("Random Error"))


def test_ssh_with_key_and_instance(
    ssh_tester, saml_client: SAMLClient, capture_command: CaptureCommand
):
    def setup(_make_stub):
        Config.touch_instance(INSTANCE_ID)
        Config.add_instance_alias(INSTANCE_ID, alias=IP_ADDRESS, region="us-east-1")
        maybe_gen_ssh_key(saml_client)

    with ssh_tester(setup=setup):
        capture_command.assert_command(
            ["exec", "true"],
            ["ssh", INSTANCE_ID],
        )


def test_ssh_with_connection_closed(
    ssh_tester,
    saml_client: SAMLClient,
    capture_command: CaptureCommand,
):
    def setup(_make_stub):
        Config.touch_instance(INSTANCE_ID)
        Config.add_instance_alias(INSTANCE_ID, alias=IP_ADDRESS, region="us-east-1")
        maybe_gen_ssh_key(saml_client)

        capture_command.register_output(
            r"ssh",
            f"Connection to {INSTANCE_ID} closed",
            exit_code=255,
        )

    with ssh_tester(setup=setup, exception=SuppressedError) as result:
        capture_command.assert_command(
            ["exec", "true"],
            ["ssh", INSTANCE_ID],
        )
        expect(result.output).to(equal(f"Connection to {INSTANCE_ID} closed\n"))
        expect(Config.check_instance_ttl(INSTANCE_ID, timedelta(minutes=1))).to(be_true)


def test_ssh_with_target_not_connected(
    ssh_tester,
    saml_client: SAMLClient,
    capture_command: CaptureCommand,
):
    def setup(_make_stub):
        Config.touch_instance(INSTANCE_ID)
        Config.add_instance_alias(INSTANCE_ID, alias=IP_ADDRESS, region="us-east-1")
        maybe_gen_ssh_key(saml_client)

        capture_command.register_output(
            r"ssh",
            f"AccessDeniedException",
            exit_code=255,
        )

    with ssh_tester(setup=setup, exception=AccessDenied) as result:
        capture_command.assert_command(
            ["exec", "true"],
            ["ssh", INSTANCE_ID],
        )
        expect(result.output).to(contain("don't have access"))
        expect(Config.check_instance_ttl(INSTANCE_ID, timedelta(minutes=1))).to(be_false)


def test_ssh_no_login(command_login_tester):
    command_login_tester(["ssh", "test", "127.0.0.1"])
    command_login_tester(["ssh", "test", "127.0.0.1", "foo", "bar", "baz"])
    command_login_tester(["ssh", "127.0.0.1"])
    command_login_tester(["ssh", "127.0.0.1"], {"SYM_RESOURCE": "test"})
    command_login_tester(["ssh", "127.0.0.1"], {"ENVIRONMENT": "test"})



import os
import pty
from typing import Sequence

from sym.cli.helpers.contexts import push_envs
from sym.cli.helpers.keywords_to_options import Argument, keywords_to_options
from sym.cli.helpers.tee import Tee
from sym.cli.saml_clients.saml_client import SAMLClient


def read_fn(client, args, *, fd_name):
    if client.log_dir:
        with Tee(client.log_dir, f"{args[0]}.{fd_name}") as tee:
            path = tee.path
        file = path.open("wb")
    else:
        file = None

    def reader(fd):
        data = os.read(fd, 1024)
        if file:
            file.write(data)
        return data

    return reader


def spawn(client: SAMLClient, args: Sequence[Argument]):
    master_read = read_fn(client, args, fd_name="stdout")
    stdin_read = read_fn(client, args, fd_name="stdin")
    with push_envs(client.get_creds()):
        pty.spawn(["bash", "-c", *keywords_to_options(args)], master_read, stdin_read)



import click

from sym.cli.actions.action_registry import ActionRegistry
from sym.cli.data.request_data import RequestData
from sym.cli.decorators import command_require_bins, loses_interactivity, require_login
from sym.cli.helpers.global_options import GlobalOptions
from sym.cli.helpers.options import resource_argument
from sym.cli.saml_clients.aws_profile import AwsCredentialsPath

from .sym import sym  # TODO: this should be an absolute import


@sym.command(short_help="Write out AWS credentials")
@resource_argument
@click.option(
    "--path",
    help="Write credentials to a specific file",
    default=str(AwsCredentialsPath),
    show_default=True,
    type=click.Path(exists=True, dir_okay=False, writable=True),
)
@click.option(
    "--profile",
    help="Profile name to write credentials to",
    default="sym",
    show_default=True,
)
@click.make_pass_decorator(GlobalOptions)
@command_require_bins("aws", "session-manager-plugin")
@require_login
@loses_interactivity
def write_creds(options: GlobalOptions, resource: str, path: str, profile: str) -> None:
    """Write out approved credentials for the specified Sym RESOURCE to
    a profile in your AWS credentials file.

    \b
    Example:
        `sym write-creds test --profile my-test-profile --path path/to/credentials`
    """
    request_data = RequestData(
        action="write_creds",
        resource=resource,
        global_options=options,
        params={"path": path, "profile": profile},
    )
    ActionRegistry.execute(request_data)



import os
from typing import ClassVar

from click import Option
from click_option_group import GroupedOption

from ..helpers.util import wrap


class Used:
    data: ClassVar[dict] = {}


def get_used():
    return Used.data


def reset_used():
    Used.data = {}


def _value_from_envvar(self, value):
    if value:
        for envvar in wrap(self.envvar):
            if os.getenv(envvar) == value:
                get_used()[envvar] = (self.name, value)
                break
    return value


class EnvvarOption(Option):
    def value_from_envvar(self, ctx):
        value = super().value_from_envvar(ctx)
        return _value_from_envvar(self, value)


class EnvvarGroupedOption(GroupedOption):
    def value_from_envvar(self, ctx):
        value = super().value_from_envvar(ctx)
        return _value_from_envvar(self, value)



import click

from . import config as cmd
from .sym import sym


def key_callback(ctx, key: str):
    return cmd.key_callback(ctx, key=f"default_{key}")


@sym.command(short_help="Get a default value")
@click.argument("key", callback=key_callback)
@click.pass_context
def defaults(ctx, key: str) -> None:
    """Get a Sym default value.

    KEY is the name of the value to get (e.g. `resource`)
    """
    ctx.forward(cmd.config)


@sym.command(name="defaults:set", short_help="Set a default value")
@click.argument("key", callback=key_callback)
@click.argument("value", callback=cmd.value_callback)
@click.pass_context
def defaults_set(ctx, key: str, value: str) -> None:
    """Set the Sym default VALUE for KEY."""
    ctx.forward(cmd.config_set)


@sym.command(name="defaults:unset", short_help="Unset a default value")
@click.argument("key", callback=key_callback)
@click.pass_context
def defaults_unset(ctx, key: str) -> None:
    """Unset the Sym default value for KEY."""
    ctx.forward(cmd.config_unset)



from dataclasses import dataclass, field
from typing import Dict, Optional, Type

from ..data.global_options_base import GlobalOptionsBase
from ..saml_clients.chooser import choose_saml_client
from ..saml_clients.saml_client import SAMLClient


@dataclass
class GlobalOptions(GlobalOptionsBase):
    saml_client_type: Type[SAMLClient] = field(
        default_factory=lambda: choose_saml_client("auto", none_ok=True)
    )
    saml_clients: Dict[str, SAMLClient] = field(default_factory=dict)
    log_dir: Optional[str] = None
    aws_region: Optional[str] = None
    disable_analytics: bool = False

    def to_dict(self):
        return {
            "debug": self.debug,
            "disable_analytics": self.disable_analytics,
            "saml_client": str(self.saml_client_type),
            "aws_region": str(self.aws_region),
        }



from typing import Tuple

import click

from ..decorators import loses_interactivity, require_login
from ..helpers.global_options import GlobalOptions
from ..helpers.options import resource_argument
from .sym import sym


@sym.command("exec", short_help="Execute a command")
@resource_argument
@click.argument("command", nargs=-1)
@click.make_pass_decorator(GlobalOptions)
@require_login
@loses_interactivity
def sym_exec(options: GlobalOptions, resource: str, command: Tuple[str, ...]) -> None:
    """Use approved creds for RESOURCE to execute COMMAND."""
    from ..saml_clients.saml_client_factory import SAMLClientFactory

    saml_client = SAMLClientFactory.create_saml_client(resource, options)
    saml_client.exec(*command)



from threading import Condition, Thread
from typing import Callable, Sequence

import click

from sym.cli.saml_clients.saml_client_factory import SAMLClientFactory

from ...errors import InstanceNotFound
from ...saml_clients.saml_client import SAMLClient
from .client import Ec2Client, RunningInstance


class MultiRegionEc2Client(Ec2Client):
    """
    A multiregion Ec2 client is able to load instances by alias from any region. Like a normal Ec2
    client, it executes describe-instances on the default region for the SAML client passed to the
    constructor. If that fails, then it calls describe-regions, and searches each region for the instance.
    """

    def __init__(self, client: SAMLClient):
        super().__init__(client)

    def _first(
        self,
        ctx,
        alias: str,
        regions: Sequence[str],
        finder: Callable[[str], RunningInstance],
        timeout=5,
    ) -> RunningInstance:
        """
        Uses the provided function to search each region and returns the first result, or
        raises InstanceNotFound after the timeout is met.

        This is pulled into a helper method to allow "ctx" to be passed in as an arg without a type
        assertion, otherwise the typechecker gets confused.
        """

        results = []
        cv = Condition()

        def search_in_region(region):
            with ctx:
                try:
                    res = finder(region)
                except InstanceNotFound:
                    return
                with cv:
                    results.append(res)
                    cv.notify()

        for region in regions:
            Thread(target=search_in_region, args=(region,), daemon=True).start()

        with cv:
            if cv.wait_for(lambda: bool(results), timeout=timeout):
                return results.pop()

        raise InstanceNotFound(alias)

    def load_instance_by_alias(self, alias: str) -> RunningInstance:
        """
        The MultiRegionEc2Client.load_instance_by_alias method first attempts
        to load the instance from the default region. If that fails with an
        InstanceNotFound error, then start a multi-threaded search for the instance
        across all regions, returning the first instance that is found or raising
        an InstanceNotFound error otherwise.
        """
        try:
            return super().load_instance_by_alias(alias)
        except InstanceNotFound:
            pass

        regions = self.get_regions()

        def find_in_region(region: str) -> RunningInstance:
            new_client = SAMLClientFactory.clone(self.client, aws_region=region)
            new_ec2_client = Ec2Client(new_client)
            return new_ec2_client.load_instance_by_alias(alias)

        ctx = click.get_current_context()
        return self._first(ctx, alias, regions, find_in_region)



from typing import Optional

from .params import PARAMS as PARAMS_BY_ORG

_ORGS_BY_DOMAIN = dict(map(lambda kv: (kv[1]["domain"], kv[0]), PARAMS_BY_ORG.items()))


def infer_org_from_email(email: str) -> Optional[str]:
    return _ORGS_BY_DOMAIN.get(email.split("@")[-1])



import datetime

import yaml


def datetime_representer(dumper, dt):
    return dumper.represent_scalar("!time", dt.isoformat())


def datetime_constructor(loader, node):
    value = loader.construct_scalar(node)
    return datetime.datetime.fromisoformat(value).time()


def sym_init():
    yaml.add_representer(datetime.time, datetime_representer)
    yaml.add_constructor("!time", datetime_constructor)



import shlex
import subprocess
from configparser import ConfigParser, NoOptionError
from pathlib import Path
from typing import Final, Iterator, Optional, Tuple

from semver import VersionInfo

from ..decorators import command_require_bins, intercept_errors, run_subprocess
from ..errors import (
    ExpiredCredentials,
    FailedSubprocessError,
    SamlClientNotSetup,
    UnavailableResourceError,
)
from ..helpers.config import Config, SymConfigFile
from ..helpers.constants import Saml2AwsCredsExpired, Saml2AwsNoCreds, Saml2AwsNoRoles
from ..helpers.contexts import push_env
from ..helpers.keywords_to_options import Argument, Options, keywords_to_options
from ..helpers.params import Profile, get_saml2aws_params
from .saml_client import SAMLClient

MIN_VERSION = VersionInfo.parse("2.26.2")

ErrorPatterns = {
    Saml2AwsNoRoles: UnavailableResourceError,
    Saml2AwsNoCreds: SamlClientNotSetup,
    Saml2AwsCredsExpired: ExpiredCredentials,
}


class Saml2Aws(SAMLClient):
    __slots__ = ["config_file", "resource", "options", "_config"]
    binary = "saml2aws"
    option_value = "saml2aws"
    priority = 10
    setup_help = "Upgrade to the latest version, then run `saml2aws login`."

    resource: str
    options: "GlobalOptionsBase"
    config_file: Final[SymConfigFile]
    _config: Optional[ConfigParser]
    _s2a_options: Final[Options]

    def __init__(self, resource: str, *, options: "GlobalOptionsBase") -> None:
        super().__init__(resource, options=options)
        self.config_file = SymConfigFile(resource=resource, file_name="saml2aws.cfg")
        self._s2a_options = {
            "verbose": self.debug,
            "config": str(self.config_file),
            "idp_account": self._section_name,
            "skip_prompt": True,
        }

    def _check_version(self) -> bool:
        # Can't use run_subprocess because saml2aws annoyingly outputs the version on stderr
        version = subprocess.run(
            ["saml2aws", "--version"],
            text=True,
            capture_output=True,
        ).stderr
        if not version:
            return False
        return VersionInfo.parse(version.strip()) >= MIN_VERSION

    def is_setup(self) -> bool:
        path = Path.home() / ".saml2aws"
        if not path.exists():
            return False

        if not self._check_version():
            return False

        config = ConfigParser()
        config.read(path)
        for section in config.sections():
            try:
                if config.get(section, "username"):
                    return True
            except NoOptionError:
                continue
        return False

    @intercept_errors(ErrorPatterns)
    @run_subprocess
    @command_require_bins(binary)
    def _exec(self, *args: str, **opts: str) -> Iterator[Tuple[Argument, ...]]:
        self.log_subprocess_event(args)
        # saml2aws exec actually joins all the arguments into a single string and
        # runs it with the shell. So we have to use shlex.join to get around that!
        reparseable = shlex.join(keywords_to_options([*args, opts]))
        with push_env("AWS_REGION", self.ensure_config()[self._section_name]["region"]):
            yield (
                "saml2aws",
                self._s2a_options,
                "exec",
                "--",
                reparseable,
            )

    @intercept_errors(ErrorPatterns, suppress=True)
    @run_subprocess
    @command_require_bins(binary)
    def _login(self, show_prompt=False, force=False):
        if show_prompt:
            options = {**self._s2a_options, "skip_prompt": False}
            args = {"username": Config.get_email()}
        else:
            options = self._s2a_options
            args = {}
        with push_env("AWS_REGION", self.ensure_config()[self._section_name]["region"]):
            # no-op if session active when force=False
            yield "saml2aws", options, "login", args, {"force": force}

    def _ensure_session(self, *, force: bool):
        try:
            self._login(silence_stderr_=not self.debug, force=force)
        except FailedSubprocessError:
            self._login(silence_stderr_=not self.debug, force=force, show_prompt=True)

    def _ensure_config(self, profile: Profile) -> ConfigParser:
        saml2aws_params = get_saml2aws_params()
        config = ConfigParser()
        config.read_dict(
            {
                self._section_name: {
                    "aws_profile": self._section_name,
                    "url": self.get_aws_saml_url(),
                    "provider": "Okta",
                    "skip_verify": "false",
                    "timeout": "0",
                    "aws_urn": "urn:amazon:webservices",
                    **saml2aws_params,
                    "role_arn": profile.arn,
                    "region": profile.region,
                }
            }
        )
        return config



import click

from ..version import __version__
from .sym import sym


@sym.command(short_help="Print the version")
def version() -> None:
    """Print the Sym CLI version."""
    click.echo(__version__)



import logging
from typing import Any, Callable, ClassVar, List

from click import Command, Context, Group
from sentry_sdk.api import push_scope

from sym.cli.helpers.global_options import GlobalOptions

from ..errors import get_active_env_vars
from ..saml_clients.chooser import choose_saml_client
from . import segment
from .tee import Tee, TeeStdErr, TeeStdOut


class AutoTagCommand(Command):
    """
    A command where each invocation sets the Sentry tag with the
    command's name automatically. Additionally, any CliErrors
    raised from the command are logged.
    """

    def invoke(self, ctx: Context) -> Any:
        segment.track(
            "Command Executed",
            global_options=ctx.find_object(GlobalOptions),
            command=ctx.info_name,
            options=ctx.obj.to_dict(),
        )
        with push_scope() as scope:
            scope.set_tag("command", ctx.info_name)
            scope.set_extra("options", ctx.obj.to_dict())
            return super().invoke(ctx)

    def parse_args(self, ctx, args):
        """
        To work around https://github.com/pallets/click/issues/714, we are doing a trick:
        we always parse a resource *option* from the root command, and then we sometimes
        rewrite the args passed to a child command before parsing the resource *argument*.
        """
        if (
            self.params
            and (resource := ctx.parent.params.get("resource"))
            and self.params[0].name == "resource"
            and (
                not args
                or not choose_saml_client(
                    ctx.parent.params["saml_client_name"]
                ).validate_resource(args[0])
            )
        ):
            args = [resource] + args
        return super().parse_args(ctx, args)

    def format_epilog(self, ctx, formatter):
        if not self.epilog:
            formatter.write(get_active_env_vars())
        super().format_epilog(ctx, formatter)


class SymGroup(Group):
    """
    A group where any defined commands automatically use
    AutoTagCommand.
    """

    tees: ClassVar[List[Tee]] = []
    instances = []

    def __init__(self, *args: Any, **attrs: Any) -> None:
        super().__init__(*args, **attrs)
        SymGroup.instances.append(self)

    def __del__(self):
        self.__class__.reset_tees()

    @classmethod
    def reset_tees(cls):
        for tee in cls.tees:
            tee.close()
        cls.tees = []

    def invoke(self, ctx: Context) -> Any:
        if (log_dir := ctx.params.get("log_dir")) :
            # Don't register exit handler so exceptions are teed.
            # Instead, __del__ will be called when the program exits.
            self.__class__.tees.extend((TeeStdOut(log_dir), TeeStdErr(log_dir)))
            logging_filename = Tee.path_for_fd(log_dir, "logging")
        else:
            logging_filename = None

        if ctx.params.get("debug"):
            logging.basicConfig(level=logging.DEBUG, filename=logging_filename)
            logging.getLogger("segment").setLevel(logging.WARNING)
        else:
            logging.getLogger("segment").setLevel(logging.CRITICAL)

        return super().invoke(ctx)

    def command(
        self, *args: Any, **kwargs: Any
    ) -> Callable[[Callable[..., Any]], AutoTagCommand]:
        return super().command(*args, **kwargs, cls=AutoTagCommand)  # type: ignore



from multiprocessing import Pool

from sym.cli.helpers.config import Config
from sym.cli.helpers.ssh import _preprocess_args, maybe_gen_ssh_key, ssh_key_and_config
from sym.cli.tests.conftest import empty_saml_client, setup_context
from sym.cli.tests.helpers.sandbox import Sandbox


def _gen_ssh_key_fn(_i=0):
    with setup_context():
        maybe_gen_ssh_key(empty_saml_client())


def test_gen_ssh_key(sandbox: Sandbox):
    with sandbox.push_xdg_config_home():
        with Pool(processes=4) as pool:
            pool.map(_gen_ssh_key_fn, range(10))

        ssh_key, _ = ssh_key_and_config(empty_saml_client())
        with ssh_key as f:
            assert "PRIVATE KEY" in f.read()


def test_gen_ssh_key_exists(sandbox: Sandbox):
    with sandbox.push_xdg_config_home():
        _gen_ssh_key_fn()
        ssh_key, _ = ssh_key_and_config(empty_saml_client())

        with ssh_key as f:
            key = f.read()

        with Pool(processes=4) as pool:
            pool.map(_gen_ssh_key_fn, range(10))

        with ssh_key as f:
            assert key == f.read()


def test_preprocess_args_invalid_controlpath_equals(sandbox: Sandbox):
    with sandbox.push_xdg_config_home():
        processed = _preprocess_args(
            empty_saml_client(),
            "fake-instance",
            ["-o", "ControlPath=/path/to/fake/cp.socket", "-o", "ControlMaster=15m"],
        )
        assert processed == ["-o", "ControlMaster=no"]


def test_preprocess_args_invalid_controlpath_space(sandbox: Sandbox):
    with sandbox.push_xdg_config_home():
        processed = _preprocess_args(
            empty_saml_client(),
            "fake-instance",
            ["-o", "ControlPath /path/to/fake/cp.socket", "-o", "ControlMaster     15m"],
        )
        assert processed == ["-o", "ControlMaster=no"]


def test_preprocess_args_valid_controlpath(sandbox: Sandbox):
    with sandbox.push_xdg_config_home():
        Config.touch_instance("fake-instance")
        processed = _preprocess_args(
            empty_saml_client(),
            "fake-instance",
            ["-o", "ControlPath /path/to/fake/cp.socket", "-o", "ControlMaster 15m"],
        )
        assert processed == [
            "-o",
            "ControlPath /path/to/fake/cp.socket",
            "-o",
            "ControlMaster 15m",
        ]


def test_preprocess_args_after_ssh_error(sandbox: Sandbox):
    with sandbox.push_xdg_config_home():
        Config.touch_instance("fake-instance", True)
        processed = _preprocess_args(
            empty_saml_client(),
            "fake-instance",
            ["-o", "ControlPath /path/to/fake/cp.socket", "-o", "ControlMaster 15m"],
        )
        assert processed == [
            "-o",
            "ControlMaster=no",
        ]



from configparser import ConfigParser
from contextlib import contextmanager
from copy import copy
from datetime import datetime, timedelta
from pathlib import Path
from typing import Dict, Optional
from unittest.mock import ANY, patch

import pytest
from expects import expect, have_keys

from sym.cli.actions.write_creds_action import WriteCredsAction
from sym.cli.data.request_data import RequestData
from sym.cli.errors import CliError
from sym.cli.helpers.config import Config
from sym.cli.helpers.global_options import GlobalOptions
from sym.cli.saml_clients.aws_okta import AwsOkta
from sym.cli.saml_clients.saml_client import SAMLClient
from sym.cli.tests.conftest import _env_str
from sym.cli.tests.helpers.capture import CaptureCommand
from sym.cli.tests.helpers.sandbox import Sandbox

TEST_PROFILE = "foobar-9000"
TEST_RESOURCE = "test"

# mapping of keys stored in .aws/credentials to internal keys
# in the case they're different, as we change some during write_creds
CRED_KEYS_MAPPING = {
    "REGION": "AWS_REGION",
    "X_SECURITY_TOKEN_EXPIRES": "AWS_CREDENTIAL_EXPIRATION",
}


class TestWriteCreds:
    """write_creds testing suite.

    Includes tests for click commands / user interaction
    and the internal module code separately.

    This class is broken into 2 distinct parts:
        1. pytest test cases
        2. pytest fixture and data creation helper methods

    To run the tests you may, from the runtime/sym-cli project folder:
    - Run the entire test suite:
        python -m pytest sym/cli/tests/commands/test_write_creds.py

    - Run a single test method:
        python -m pytest sym/cli/tests/commands/test_write_creds.py::TestWriteCreds::test_execute_write_creds
    """

    def test_execute_write_creds(
        self, sandbox, credentials_path, creds_env, capture_command
    ):
        with self.write_creds_context_manager(capture_command, sandbox, creds_env):
            global_options = GlobalOptions(saml_client_type=AwsOkta)

            request_data = RequestData(
                action="write_creds",
                resource=TEST_RESOURCE,
                global_options=global_options,
                params={"path": str(credentials_path), "profile": TEST_PROFILE},
            )
            action = WriteCredsAction()
            action.execute(request_data)

            written_creds = self.get_credentials_from_file(credentials_path, TEST_PROFILE)
            expect(written_creds).to(have_keys(creds_env))

    def test_execute_write_creds_without_aws_okta_fails(
        self, sandbox, credentials_path, capture_command, creds_env
    ):
        with pytest.raises(CliError, match="Unable to find aws-okta in your path!"):
            with self.write_creds_context_manager(
                capture_command, sandbox, create_binary=False
            ):
                global_options = GlobalOptions(saml_client_type=AwsOkta)

                request_data = RequestData(
                    action="write_creds",
                    resource=TEST_RESOURCE,
                    global_options=global_options,
                    params={"path": str(credentials_path), "profile": TEST_PROFILE},
                )
                action = WriteCredsAction()
                action.execute(request_data)

    @patch("sym.cli.saml_clients.saml_client.SAMLClient._ensure_session")
    def test_execute_write_creds_expiring_force_ensures_session(
        self, _ensure_session, sandbox, credentials_path, capture_command, creds_env
    ):
        creds = copy(creds_env)
        creds["AWS_OKTA_SESSION_EXPIRATION"] = str(int(datetime.now().timestamp()))

        with self.write_creds_context_manager(capture_command, sandbox, creds):
            global_options = GlobalOptions(saml_client_type=AwsOkta)

            request_data = RequestData(
                action="write_creds",
                resource=TEST_RESOURCE,
                global_options=global_options,
                params={"path": str(credentials_path), "profile": TEST_PROFILE},
            )

            action = WriteCredsAction()
            action.execute(request_data)

            _ensure_session.assert_called_once_with(force=True)

            written_creds = self.get_credentials_from_file(credentials_path, TEST_PROFILE)

            # Ideally we'd like to make sure the written_creds expiration is in
            # the future, not what we set as creds["AWS_OKTA_SESSION_EXPIRATION"],
            # but since we stub the subprocess, we can't get updated creds.
            # So the real test here is _ensure_session.assert_called and this check
            # is extra.
            expect(written_creds).to(have_keys(creds))

    @patch(
        "sym.cli.saml_clients.saml_client.SAMLClient._profile_matches_caller_identity",
        return_value=False,
    )
    @patch("sym.cli.saml_clients.saml_client.SAMLClient._ensure_session")
    def test_execute_write_creds_profile_mismatch_force_ensures_session(
        self,
        _ensure_session,
        _profile_match,
        sandbox,
        credentials_path,
        capture_command,
        creds_env,
    ):
        creds = copy(creds_env)

        # _profile_matches_caller_identity doesn't get called if creds are expiring
        future_datetime = datetime.now() + timedelta(minutes=30)
        creds["AWS_OKTA_SESSION_EXPIRATION"] = str(int(future_datetime.timestamp()))

        with self.write_creds_context_manager(capture_command, sandbox, creds):
            global_options = GlobalOptions(saml_client_type=AwsOkta)

            request_data = RequestData(
                action="write_creds",
                resource=TEST_RESOURCE,
                global_options=global_options,
                params={"path": str(credentials_path), "profile": TEST_PROFILE},
            )

            action = WriteCredsAction()
            action.execute(request_data)

            _profile_match.assert_called_once()
            _ensure_session.assert_called_once_with(force=True)

            written_creds = self.get_credentials_from_file(credentials_path, TEST_PROFILE)

            expect(written_creds).to(have_keys(creds))

    def test_write_creds_no_login_errors(self, command_login_tester):
        command_login_tester(["write-creds"])
        command_login_tester(["write-creds", TEST_RESOURCE])
        command_login_tester(["write-creds"], {"SYM_RESOURCE": TEST_RESOURCE})
        command_login_tester(["write-creds"], {"ENVIRONMENT": TEST_RESOURCE})

    def test_write_creds_command_executes_correct_action_with_params(
        self, simple_command_tester, capture_command, creds_env, credentials_path, sandbox
    ):
        with self.write_creds_context_manager(capture_command, sandbox, creds_env):
            sandbox.create_binary(f"bin/aws")
            sandbox.create_binary(f"bin/session-manager-plugin")

            with patch.object(WriteCredsAction, "write_creds") as mock_write_creds:
                result = simple_command_tester(
                    [
                        "write-creds",
                        TEST_RESOURCE,
                        "--profile",
                        TEST_PROFILE,
                        "--path",
                        str(credentials_path),
                    ],
                )

                assert result.exit_code == 0
                mock_write_creds.assert_called_once_with(
                    TEST_RESOURCE, ANY, str(credentials_path), TEST_PROFILE
                )

    ##############################################################################################
    # SECTION 2
    # pytest fixture and testing data preparation
    # and data creation helper methods
    #
    ##############################################################################################

    @pytest.fixture
    def credentials_path(self, sandbox: Sandbox):
        """Create `.aws/credentials` file in the sandbox environment
        and return the path.
        """

        credentials_path = sandbox.path / ".aws" / "credentials"

        try:
            credentials_path.parent.mkdir(parents=True)
            credentials_path.touch()
        except FileExistsError:
            # Allow tests to use this fixture to get the path even
            # if they've already used this fixture to create the file.
            pass

        return credentials_path

    def get_credentials_from_file(self, path: Path, profile: str) -> dict:
        """Read AWS credentials file at path and return all creds
        for the provided profile as a dict.

        Does some replacements to reverse changes we make when writing
        credentials to the file so it can be compared to internal dicts
        like the creds_env fixture easily.
        """

        config = ConfigParser()
        with open(path) as f:
            config.read_file(f)

        creds = config.items(profile)
        creds_dict = {k.upper(): v for k, v in creds}

        for file_key_name, internal_key_name in CRED_KEYS_MAPPING.items():
            val = creds_dict.pop(file_key_name)
            creds_dict[internal_key_name] = val

        return creds_dict

    @contextmanager
    def write_creds_context_manager(
        self,
        capture_command: CaptureCommand,
        sandbox: Sandbox,
        creds_env: Optional[Dict[str, str]] = None,
        create_binary: bool = True,
    ):
        """Load all required context for executing WriteCredsAction.

        Initializes capture_command for stubbing subprocesses,
        stubs the process to get creds from the environment,
        simulates a sym login with the "sym" org and "y@symops.io" email,
        and initializes a sandbox with an AwsOkta binary.

        See test_execute_write_creds as an example of usage.

        Args:
            capture_command: instance of CaptureCommand to stub subprocesses
            sandbox: instance of Sandbox to keep tests isolated
            creds_env: key/value pairs to set as the result for the subprocess
                that retrieves creds from the local environment
            create_binary: whether to create the AwsOkta binary
        """

        if not creds_env:
            creds_env = {}

        with capture_command():
            capture_command.register_output(r"env", _env_str(creds_env))

            with sandbox.push_xdg_config_home():
                Config.instance()["org"] = "sym"
                Config.instance()["email"] = "y@symops.io"

                if create_binary:
                    sandbox.create_binary(f"bin/{AwsOkta.binary}")

                with sandbox.push_exec_path():
                    yield



import hashlib
import json
import os
from contextlib import contextmanager
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import ClassVar

from portalocker.exceptions import AlreadyLocked, LockException

from ..io import TruncatingStringIO
from ..os import create_lock, read_lock, read_write_lock


class _Config:
    dir_name: ClassVar[str]


def init(dir_name):
    _Config.dir_name = dir_name


def xdg_config_home() -> Path:
    try:
        return Path(os.environ["XDG_CONFIG_HOME"]).expanduser().resolve()
    except KeyError:
        return Path.home() / ".config"


def sym_config_file(file_name: str) -> Path:
    return (xdg_config_home() / _Config.dir_name / file_name).resolve()


class SymConfigFile:
    def __init__(self, *, file_name: str, uid_scope: bool = True, **dependencies):
        path = Path()

        if dependencies:
            key = json.dumps(dependencies, sort_keys=True).encode("utf-8")
            md5 = hashlib.md5(key).hexdigest()
            path = path / md5

        path = path / file_name

        if uid_scope:
            path = Path("uid") / str(os.getuid()) / path
        else:
            path = Path("default") / path

        self.path = sym_config_file(path)

        self.read_lock = read_lock(self.path)
        self.write_lock = create_lock(self.path)
        self.update_lock = read_write_lock(self.path)

    def __enter__(self):
        self.mkparents()
        try:
            # acquire read lock, hold across context
            fh = self.read_lock.acquire()
            fh.seek(0)  # reentrant
            self.value = fh.read()
        except FileNotFoundError:
            self.value = None
        self.file = TruncatingStringIO(initial_value=self.value)
        return self.file

    def __exit__(self, type, value, traceback):
        self.read_lock.release()  # release read lock
        if value:
            return
        if self.file.tell() == 0:
            return
        if self.value != (value := self.file.getvalue()):
            with self.exclusive_access() as f:
                f.write(value)

    def __str__(self):
        return str(self.path)

    def mkparents(self):
        self.path.parent.mkdir(parents=True, exist_ok=True)

    def put(self, s: str):
        with self as f:
            f.write(s)

    def read(self):
        if self.value:
            return self.value
        with self.read_lock.acquire() as fh:
            fh.seek(0)
            return fh.read()

    def _throwaway_write(self):
        with NamedTemporaryFile(mode="w+") as f:
            yield f

    @contextmanager
    def exclusive_access(self):
        try:
            with self.write_lock as f:  # acquire write lock
                yield f
        except AlreadyLocked:  # another thread is writing same value
            yield from self._throwaway_write()

    @contextmanager
    def exclusive_create(self):
        self.mkparents()
        if self.path.exists():
            yield from self._throwaway_write()
        else:
            with self.exclusive_access() as f:
                yield f

    @contextmanager
    def update(self, fail_ok=False):
        self.mkparents()
        try:
            with self.update_lock as f:
                f.seek(0)  # reentrant
                yield f
        except LockException:
            if fail_ok:
                yield TruncatingStringIO(initial_value=self.read())
            else:
                raise



from unittest.mock import ANY, patch

import pytest

from sym.cli.decorators import setup_sentry
from sym.cli.helpers.config import Config


class TestSetupSentry:
    @pytest.mark.parametrize(
        argnames=("is_prod_mode", "org_name", "expected_environment"),
        argvalues=[
            (True, "sym", "development"),
            (True, "notsym", "pypi"),
            (False, "sym", "development"),
            (False, "notsym", "development"),
        ],
    )
    @patch("sym.cli.decorators.sentry_init")
    def test_sentry_environment(
        self,
        mock_sentry_init,
        is_prod_mode: bool,
        org_name: str,
        expected_environment: str,
        sandbox,
    ):
        with patch("sym.cli.decorators.is_prod_mode", return_value=is_prod_mode):
            with sandbox.push_xdg_config_home():
                Config.instance()["org"] = org_name
                setup_sentry()
                mock_sentry_init.assert_called_once_with(
                    environment=expected_environment, sample_rate=ANY
                )



from multiprocessing import Pool

import pytest

from sym.cli.saml_clients.aws_okta import AwsOkta
from sym.cli.tests.helpers.capture import CaptureCommand
from sym.cli.tests.saml_clients.conftest import TestContextFixture

pytestmark = [
    pytest.mark.usefixtures("click_context"),
    pytest.mark.parametrize(argnames=["constructor"], argvalues=[[AwsOkta]]),
]


def test_aws_okta(
    test_context: TestContextFixture[AwsOkta], capture_command: CaptureCommand
) -> None:
    with test_context(debug=False) as aws_okta:
        with capture_command():
            aws_okta.exec("aws", "ssm", "start-session", target="i-0123456789abcdef")
    capture_command.assert_command(
        "aws-okta exec sym-catamorphic -- env",
        "aws-okta --debug exec sym-catamorphic -- false",
        "aws-okta exec sym-catamorphic -- true",
        "aws-okta exec sym-catamorphic -- aws ssm start-session --target i-0123456789abcdef",
    )


def test_aws_okta_debug(
    test_context: TestContextFixture[AwsOkta], capture_command: CaptureCommand
) -> None:
    with test_context(debug=True) as aws_okta:
        with capture_command():
            aws_okta.exec("env")
    capture_command.assert_command(
        "aws-okta --debug exec sym-catamorphic -- env",
        "aws-okta --debug exec sym-catamorphic -- false",
        "aws-okta --debug exec sym-catamorphic -- true",
        "aws-okta --debug exec sym-catamorphic -- env",
    )


def test_aws_okta_multi_threads(subprocess_friendly_asserter) -> None:
    commands = (
        "aws-okta exec sym-test -- env",
        "aws-okta --debug exec sym-test -- false",
        "aws-okta exec sym-test -- true",
        "aws-okta exec sym-test -- env",
    )
    f, *args = subprocess_friendly_asserter
    with Pool(processes=4) as pool:
        pool.map(f, [(*args, commands)] * 100)



from getpass import getuser
from os import uname
from typing import Optional
from uuid import uuid4

import analytics

from ..version import __version__
from .config import Config


def _context():
    u = uname()
    return {
        "app": {"name": "sym-cli", "version": __version__},
        "os": {"name": u.sysname, "version": u.release},
    }


def _user_id():
    return f"{Config.get_org()}:{Config.get_email()}"


def _identity_kwargs():
    try:
        return {"user_id": _user_id(), "context": _context()}
    except KeyError:
        return {"anonymous_id": str(uuid4()), "context": _context()}


def _skip(global_options: Optional["GlobalOptions"]):
    if not Config.is_logged_in():
        return True
    if global_options:
        return global_options.disable_analytics

    return False


def identify(global_options: Optional["GlobalOptions"] = None):
    if _skip(global_options):
        return

    traits = {"username": getuser()}

    for attr in ("org", "email"):
        try:
            traits[attr] = Config.instance()[attr]
        except KeyError:
            pass

    kwargs = _identity_kwargs()
    analytics.identify(traits=traits, **kwargs)
    if "user_id" in kwargs:
        analytics.group(group_id=Config.get_org(), **kwargs)


def track(event, global_options: Optional["GlobalOptions"] = None, **properties):
    if _skip(global_options):
        return

    analytics.track(event=event, properties=properties, **_identity_kwargs())



from contextlib import contextmanager
from functools import wraps
from typing import (
    Any,
    ClassVar,
    Final,
    Generic,
    Iterator,
    MutableMapping,
    Optional,
    TypeVar,
    cast,
)

import immutables
import yaml

from .sym_config_file import SymConfigFile


def fail_ok(f):
    @wraps(f)
    def wrapper(cls, *args, **kwargs):
        with cls.fail_ok():
            return f(cls, *args, **kwargs)

    return wrapper


ConfigSchema = TypeVar("ConfigSchema")


class ConfigBase(Generic[ConfigSchema], MutableMapping[str, Any]):
    __slots__ = ["file", "config"]

    __fail_ok: ClassVar[bool] = False

    file: Final[SymConfigFile]
    config: Final[ConfigSchema]

    def __init__(self) -> None:
        self.file = SymConfigFile(file_name="config.yml", uid_scope=False)
        with self.file as f:
            self.__load(f)

    def __load(self, fh):
        self.config = cast(ConfigSchema, yaml.safe_load(stream=fh) or {})

    def __flush(self, fh) -> None:
        fh.seek(0)
        fh.truncate()
        yaml.safe_dump(self.config, stream=fh)

    def __getitem__(self, key: str) -> Any:
        item = self.config[key]
        if isinstance(item, dict):
            return immutables.Map(item)
        return item

    def __delitem__(self, key: str) -> None:
        with self.atomic() as f:
            del self.config[key]
            self.__flush(f)

    def __setitem__(self, key: str, value: Any) -> None:
        with self.atomic() as f:
            if isinstance(value, immutables.Map):
                value = dict(value)
            self.config[key] = value
            self.__flush(f)

    def __iter__(self) -> Iterator[str]:
        return cast(Iterator[str], iter(self.config))

    def __len__(self) -> int:
        return len(self.config)

    def __repr__(self) -> str:
        return f"{type(self).__name__}({str(self.file)})"

    @classmethod
    @contextmanager
    def fail_ok(cls):
        cls.__fail_ok = True
        yield
        cls.__fail_ok = False

    @contextmanager
    def atomic(self):
        with self.file.update(fail_ok=self.__class__.__fail_ok) as f:
            self.__load(f)
            yield f

    @classmethod
    def reset(cls):
        setattr(cls, "__instance", cls())

    @classmethod
    def instance(cls) -> "ConfigBase":
        if not hasattr(cls, "__instance"):
            cls.reset()
        return getattr(cls, "__instance")

    @classmethod
    def get_default(cls, key) -> Optional[str]:
        return cls.instance().get(f"default_{key}")

    @classmethod
    def is_logged_in(cls):
        return "email" in cls.instance()

    @classmethod
    def get_email(cls) -> str:
        return cls.instance()["email"]



import pytest

from sym.cli.helpers.global_options import GlobalOptions
from sym.cli.saml_clients.saml_client import SAMLClient

TEST_RESOURCE = "foobar"


class BasicTestClient(SAMLClient):
    def _ensure_config(self, profile: "Profile") -> "ConfigParser":
        pass

    def is_setup(self) -> bool:
        return True

    def _exec(self, *args: str, **opts: str) -> None:
        pass


class TestSAMLClient:
    """SAMLClient abstract class testing suite. Uses the BasicSAMLClient class
    to test the base methods.

    To run the tests you may, from the runtime/sym-cli project folder:
    - Run the entire test suite:
        python -m pytest sym/cli/tests/saml_clients/test_saml_client.py

    - Run a single test method:
        python -m pytest sym/cli/tests/saml_clients/test_saml_client.py::TestSAMLClient::test_missing_arn_errors
    """

    def test_testing_client_not_included_in_sorted_subclasses(self):
        subclasses = SAMLClient.sorted_subclasses()
        assert subclasses
        assert BasicTestClient not in subclasses

    def test_missing_arn_errors(self):
        with pytest.raises(ValueError) as exception_info:
            client = BasicTestClient(resource=TEST_RESOURCE, options=GlobalOptions())
            client.resource_from_arn(None)

        assert exception_info.match("missing ARN")



import socket

import pytest

from sym.cli.errors import InstanceNotFound
from sym.cli.helpers.ec2.client import Ec2Client, InstanceQuery, InvalidQueryError
from sym.cli.tests.conftest import MonkeyPatch
from sym.cli.tests.helpers.ec2.conftest import (
    BOTO_EMPTY_RESPONSE,
    BOTO_INSTANCE_RESPONSE,
    TEST_DNS_NAME,
    TEST_HOSTNAME,
    TEST_INSTANCE,
    TEST_INSTANCE_ID,
    TEST_IP,
    TEST_REGION,
)


def test_get_regions(ec2_stub, ec2_client):
    ec2_stub.add_response(
        "describe_regions",
        {"Regions": [{"RegionName": TEST_REGION}]},
    )
    regions = ec2_client.get_regions()
    assert regions == [TEST_REGION]


def test_get_regions_boto_error_fallback(ec2_stub, ec2_client: Ec2Client):
    ec2_stub.add_client_error(
        "describe_regions",
        service_message="Error describing regions",
        http_status_code=400,
    )
    regions = ec2_client.get_regions()
    assert len(regions) == 20
    assert "us-east-1" in regions


def test_load_instance_by_id(ec2_stub, ec2_client: Ec2Client):
    ec2_stub.add_response(
        "describe_instances",
        BOTO_INSTANCE_RESPONSE,
        InstanceQuery().instance_id(TEST_INSTANCE_ID).params(),
    )
    instance = ec2_client.load_instance_by_alias(TEST_INSTANCE_ID)
    assert instance == TEST_INSTANCE


def test_load_instance_by_id_boto_error(ec2_stub, ec2_client: Ec2Client):
    with pytest.raises(InstanceNotFound, match=TEST_INSTANCE_ID):
        ec2_stub.add_client_error(
            "describe_instances",
            service_message="Error describing instances",
            service_error_code="InvalidInstanceID.NotFound",
            http_status_code=400,
        )
        ec2_client.load_instance_by_alias(TEST_INSTANCE_ID)


def test_load_instance_by_id_no_res(ec2_stub, ec2_client: Ec2Client):
    with pytest.raises(InstanceNotFound, match=TEST_INSTANCE_ID):
        ec2_stub.add_response(
            "describe_instances",
            {"Reservations": []},
            InstanceQuery().instance_id(TEST_INSTANCE_ID).params(),
        )
        ec2_client.load_instance_by_alias(TEST_INSTANCE_ID)


def test_load_instance_by_id_no_inst(ec2_stub, ec2_client: Ec2Client):
    with pytest.raises(InstanceNotFound, match=TEST_INSTANCE_ID):
        ec2_stub.add_response(
            "describe_instances",
            {"Reservations": [{"Instances": []}]},
            InstanceQuery().instance_id(TEST_INSTANCE_ID).params(),
        )
        ec2_client.load_instance_by_alias(TEST_INSTANCE_ID)


def test_load_instance_by_id_too_many_inst(ec2_stub, ec2_client: Ec2Client):
    with pytest.raises(InvalidQueryError, match=TEST_INSTANCE_ID):
        ec2_stub.add_response(
            "describe_instances",
            {
                "Reservations": [
                    {
                        "Instances": [
                            {"InstanceId": TEST_INSTANCE_ID},
                            {"InstanceId": "foo"},
                        ]
                    }
                ]
            },
            InstanceQuery().instance_id(TEST_INSTANCE_ID).params(),
        )
        ec2_client.load_instance_by_alias(TEST_INSTANCE_ID)


def test_load_instance_by_private_ip(ec2_stub, ec2_client: Ec2Client):
    ec2_stub.add_response(
        "describe_instances",
        BOTO_INSTANCE_RESPONSE,
        InstanceQuery().filter("ip-address", [TEST_IP]).params(),
    )
    instance = ec2_client.load_instance_by_alias(TEST_IP)
    assert instance == TEST_INSTANCE


def test_load_instance_by_private_ip_no_inst(ec2_stub, ec2_client: Ec2Client):
    with pytest.raises(InstanceNotFound, match=TEST_IP):
        ec2_stub.add_response(
            "describe_instances",
            BOTO_EMPTY_RESPONSE,
            InstanceQuery().filter("ip-address", [TEST_IP]).params(),
        )
        ec2_stub.add_response(
            "describe_instances",
            BOTO_EMPTY_RESPONSE,
            InstanceQuery().filter("private-ip-address", [TEST_IP]).params(),
        )
        ec2_client.load_instance_by_alias(TEST_IP)


def test_load_instance_by_private_dns(ec2_stub, ec2_client: Ec2Client):
    ec2_stub.add_response(
        "describe_instances",
        BOTO_INSTANCE_RESPONSE,
        InstanceQuery().filter("dns-name", [TEST_DNS_NAME]).params(),
    )
    instance = ec2_client.load_instance_by_alias(TEST_DNS_NAME)
    assert instance == TEST_INSTANCE


def test_load_instance_by_dns_socket_error(
    ec2_stub, ec2_client: Ec2Client, monkeypatch: MonkeyPatch
):
    mock = {"called": False}

    def get_host(_: str) -> str:
        mock["called"] = True
        raise socket.gaierror(1, "")

    monkeypatch.setattr(socket, "gethostbyname", get_host)

    with pytest.raises(InstanceNotFound, match=TEST_HOSTNAME):
        ec2_stub.add_response(
            "describe_instances",
            BOTO_EMPTY_RESPONSE,
            InstanceQuery().filter("dns-name", [TEST_HOSTNAME]).params(),
        )
        ec2_stub.add_response(
            "describe_instances",
            BOTO_EMPTY_RESPONSE,
            InstanceQuery().filter("private-dns-name", [TEST_HOSTNAME]).params(),
        )
        ec2_client.load_instance_by_alias(TEST_HOSTNAME)

    assert mock["called"]


def test_load_instance_by_dns_lookup_then_ip_unknown(
    ec2_stub, ec2_client: Ec2Client, monkeypatch: MonkeyPatch
):
    def get_host(_: str) -> str:
        return TEST_IP

    monkeypatch.setattr(socket, "gethostbyname", get_host)

    with pytest.raises(InstanceNotFound, match=TEST_HOSTNAME):
        ec2_stub.add_response(
            "describe_instances",
            BOTO_EMPTY_RESPONSE,
            InstanceQuery().filter("dns-name", [TEST_HOSTNAME]).params(),
        )
        ec2_stub.add_response(
            "describe_instances",
            BOTO_EMPTY_RESPONSE,
            InstanceQuery().filter("private-dns-name", [TEST_HOSTNAME]).params(),
        )
        ec2_stub.add_response(
            "describe_instances",
            BOTO_EMPTY_RESPONSE,
            InstanceQuery().filter("ip-address", [TEST_IP]).params(),
        )
        ec2_stub.add_response(
            "describe_instances",
            BOTO_EMPTY_RESPONSE,
            InstanceQuery().filter("private-ip-address", [TEST_IP]).params(),
        )
        ec2_client.load_instance_by_alias(TEST_HOSTNAME)


def test_load_instance_by_dns_lookup_then_ip_success(ec2_stub, ec2_client: Ec2Client):
    ec2_stub.add_response(
        "describe_instances",
        BOTO_EMPTY_RESPONSE,
        InstanceQuery().filter("dns-name", ["localhost"]).params(),
    )
    ec2_stub.add_response(
        "describe_instances",
        BOTO_EMPTY_RESPONSE,
        InstanceQuery().filter("private-dns-name", ["localhost"]).params(),
    )
    ec2_stub.add_response(
        "describe_instances",
        BOTO_INSTANCE_RESPONSE,
        InstanceQuery().filter("ip-address", ["127.0.0.1"]).params(),
    )
    instance = ec2_client.load_instance_by_alias("localhost")
    assert instance == TEST_INSTANCE



from abc import ABC, abstractmethod

from sym.cli.data.request_data import RequestData
from sym.cli.errors import CliError
from sym.cli.helpers.check.model import SymCheck


class SymAction(ABC):
    def __init__(self, name):
        self.name = name

    def validate_inputs(self, request_data: RequestData) -> None:
        """Entry point to validate request data.

        Args:
            request_data: Operational request to validate before execution

        Raises:
            ValueError: if any part of request data is invalid (e.g. missing resource)
        """

        if not request_data.resource:
            raise CliError("must specify a resource.")

    def get_requirements(self) -> [SymCheck]:
        return []

    def invoke_requirements(self):
        requirements = self.get_requirements()
        failures = []
        failure_string = ""
        if requirements:
            for requirement in requirements:
                result = requirement.check(None)
                if not result.success:
                    failures.append(result)
                    failure_string = (
                        failure_string + f"Validation Error, message:{result.msg}\n"
                    )

        if len(failures) > 0:
            raise CliError(failures)

    @abstractmethod
    def execute(self, request_data: RequestData):
        pass



from sym.cli.helpers import segment
from sym.cli.helpers.global_options import GlobalOptions

from ..errors import SAMLClientNotFound
from .saml_client import SAMLClient


class SAMLClientFactory:
    @classmethod
    def create_saml_client(cls, resource: str, options: GlobalOptions) -> SAMLClient:
        """Returns an initialized SAML Client, either from a cache based on the
        resource name, or by initializing a new one based on the type specified
        in the options provided.
        """
        if not options.saml_client_type:
            raise SAMLClientNotFound()

        segment.track("Resource Requested", global_options=options, resource=resource)
        saml_client = options.saml_clients.get(resource, None)

        if saml_client is None:
            saml_client = options.saml_client_type(resource, options=options)
            options.saml_clients[resource] = saml_client

        return saml_client

    @classmethod
    def clone(cls, client: SAMLClient, **option_overrides) -> SAMLClient:
        """Create a new SAMLClient with the same setup as the provided
        client, with option overrides from **kwargs.

        Args:
            client: SAMLClient to clone
            option_overrides: key/value pairs to override in client options
        """

        return client.clone(options=client.options.clone(**option_overrides))



import os
import re
import shlex
import subprocess
from contextlib import contextmanager
from typing import Any, Dict, Iterator, List, Pattern, Sequence, Tuple, Union

from _pytest.monkeypatch import MonkeyPatch
from expects import *


class ExecutedCommand:
    def __init__(
        self,
        args: Sequence[str],
        result: subprocess.CompletedProcess,
        env: Dict[str, str] = {},
    ):
        if isinstance(args, str):
            self.args = shlex.split(args)
        else:
            self.args = list(args)

        self.env = env
        self.result = result


class ExpectedCommand:
    def __init__(
        self,
        args: Sequence[str],
        *,
        env: Dict[str, str] = {},
    ):
        self.args = args
        self.env = env

    def assert_command(self, actual: ExecutedCommand) -> None:
        expect(actual.args).to(contain(*self.args))
        expect(actual.env.items()).to(contain(*self.env.items()))


def c(*args, **kwargs) -> ExpectedCommand:
    return ExpectedCommand(args, env=kwargs)


class CaptureCommand:
    def __init__(self, monkeypatch: MonkeyPatch):
        self.monkeypatch = monkeypatch
        self.commands: List[List[str]] = []
        self.executed_commands: List[ExecutedCommand] = []
        self.outputs = []
        self.matched_outputs: List[Tuple[Pattern, str, int, bool]] = []
        self.allowed_calls = []
        self._run = getattr(subprocess, "run")

    def _get_output(self, args: Sequence[str]):
        command = shlex.join(args)
        for (pattern, output, exit_code, stderr) in self.matched_outputs:
            if pattern.search(command):
                if exit_code:
                    raise subprocess.CalledProcessError(exit_code, args, stderr=output)
                elif stderr:
                    return subprocess.CompletedProcess(
                        args=args, returncode=0, stderr=output
                    )
                else:
                    return output
        if self.outputs:
            return self.outputs.pop(0)

    # This function is designed for capturing the run_subprocess decorator only.
    # It is not intended to capture subprocess.run calls generally. For example,
    # `args` has a much more constrained type than subprocess.run normally allows.
    def _run_stub(
        self, args: Sequence[str], **kwargs: Any
    ) -> subprocess.CompletedProcess:
        if isinstance(args, str):
            self.commands.append(shlex.split(args))
        else:
            self.commands.append(list(args))
        if args[0] in self.allowed_calls:
            output = self._run(args, **kwargs)
        else:
            output = self._get_output(args)
            if not isinstance(output, subprocess.CompletedProcess):
                output = subprocess.CompletedProcess(
                    args=args, returncode=0, stdout=output
                )
        env = dict(os.environ)
        self.executed_commands.append(ExecutedCommand(args, output, env))
        return output

    @contextmanager
    def __call__(self) -> Iterator[None]:
        with self.monkeypatch.context() as mp:
            mp.setattr(subprocess, "run", self._run_stub)
            yield

    def enqueue_outputs(self, *outputs: Sequence[str]) -> None:
        self.outputs.extend(outputs)

    def register_output(
        self, pattern: str, output: str, *, exit_code=0, stderr: bool = False
    ) -> None:
        """Stubs a subprocess matching the provided pattern and sets its output."""
        self.matched_outputs.append((re.compile(pattern), output, exit_code, stderr))

    def allow_call(self, binary: str) -> None:
        self.allowed_calls.append(binary)

    def assert_command(self, *commands: Union[str, Sequence[str]]) -> None:
        expect(self.commands).to(have_len(len(commands)))
        for (actual, expected) in zip(self.commands, commands):
            if isinstance(expected, Pattern):
                expect(shlex.join(actual)).to(match(expected))
            elif isinstance(expected, list):
                expect(actual).to(contain(*expected))
            else:
                expect(actual).to(equal(shlex.split(expected)))

    def assert_commands(self, commands: List[ExpectedCommand]) -> None:
        expect(self.executed_commands).to(have_len(len(commands)))
        for (actual, expected) in zip(self.executed_commands, commands):
            expected.assert_command(actual)



import os
import shlex
import subprocess
import sys
import time
from functools import wraps
from subprocess import PIPE, CalledProcessError
from typing import Any, Callable, Iterator, Optional, Sequence, TypeVar, cast

import analytics
import click
from sentry_sdk.api import configure_scope
from sentry_sdk.hub import init as sentry_init

from .errors import (
    CliError,
    ErrorPatterns,
    FailedSubprocessError,
    SuppressedError,
    raise_if_match,
)
from .helpers import segment
from .helpers.config import Config
from .helpers.global_options import GlobalOptions
from .helpers.keywords_to_options import Argument, keywords_to_options
from .helpers.os import has_command
from .helpers.sentry import set_scope_context_os, set_scope_tag_org, set_scope_user
from .helpers.tee import Tee
from .saml_clients.saml_client_factory import SAMLClientFactory

# Typing support for decorators is still in need of a lot of work:
# https://github.com/python/mypy/issues/3157. So the current grossness
# with casts is unavoidable for now.


F = TypeVar("F", bound=Callable[..., Any])


def is_prod_mode():
    return os.getenv("SENTRY_ENVIRONMENT") != "development" and os.getenv("CI") != "true"


def run_subprocess(
    func: Callable[..., Iterator[Sequence[Argument]]]
) -> Callable[..., Optional[Sequence[Optional[str]]]]:
    @wraps(func)
    def impl(
        *args: Any,
        censor_: bool = False,
        capture_output_: bool = False,
        silence_stderr_: bool = True,
        input_: str = None,
        run_subprocess_options_: Optional[GlobalOptions] = None,
        **kwargs: Any,
    ) -> Optional[Sequence[Optional[str]]]:

        options = run_subprocess_options_
        if not options:
            # Ideally we will never need to access click at this level and
            # will instead always pass in options. This is for backward
            # compatibility as we refactor.
            options = click.get_current_context().find_object(GlobalOptions)

        outputs = []
        tee = bool(options.log_dir) and not censor_

        for command in func(*args, **kwargs):
            command = keywords_to_options(command)
            options.dprint(f"exec: {shlex.join(command)}\n")
            if tee:
                command = Tee.tee_command(options.log_dir, command)
            result = subprocess.run(
                command,
                check=True,
                capture_output=capture_output_,
                input=input_,
                stderr=None if (capture_output_ or not silence_stderr_) else PIPE,
                text=True,
                shell=tee,
                executable="/bin/bash" if tee else None,
            )
            outputs.append(result.stdout)
            if not censor_:
                options.dprint(result.stdout)
        if capture_output_:
            return outputs

    return impl


def intercept_errors(
    patterns: ErrorPatterns = {},
    *,
    quiet: bool = False,
    suppress: bool = False,
) -> Callable[[F], F]:
    def decorate(fn: F) -> F:
        @wraps(fn)
        def wrapped(
            *args: Any,
            quiet_: bool = False,
            suppress_: bool = False,
            intercept_errors_options_: Optional[GlobalOptions] = None,
            **kwargs: Any,
        ) -> Any:

            options = intercept_errors_options_
            if not options:
                # Ideally we will never need to access click at this level and
                # will instead always pass in options. This is for backward
                # compatibility as we refactor.
                options = click.get_current_context().find_object(GlobalOptions)

            try:
                return fn(*args, **kwargs)
            except CalledProcessError as err:
                message = err.stderr or ""
                raise_if_match(patterns, message)
                if not (quiet or quiet_):
                    sys.stderr.write(message)
                else:
                    options.dprint(message)
                if suppress or suppress_:
                    raise SuppressedError(err) from err
                else:
                    raise FailedSubprocessError(err) from err

        return cast(F, wrapped)

    return decorate


def command_require_bins(*bins: str) -> Callable[[F], F]:
    """Decorator to declare that the Click command being defined requires
    specific binaries to run successfully.

    See: https://click.palletsprojects.com/en/7.x/commands/#decorating-commands

    Raises:
        CliError: if any required bin could not be found
    """

    def decorate(fn: F) -> F:
        @click.pass_context
        @wraps(fn)
        def wrapped(context: click.Context, *args: Any, **kwargs: Any) -> Any:
            for binary in bins:
                if not has_command(binary):
                    raise CliError(f"Unable to find {binary} in your path!")
            return context.invoke(fn, *args, **kwargs)

        return cast(F, wrapped)

    return decorate


def require_bins(*bins: str) -> Callable[[F], F]:
    """Decorator to declare that the function being defined requires
    specific binaries to run successfully.

    Use this decorator to require bins on functions without Click context.

    Raises:
        CliError: if any required bin could not be found
    """

    def decorate(fn: F) -> F:
        @wraps(fn)
        def wrapped(*args: Any, **kwargs: Any) -> Any:
            for binary in bins:
                if not has_command(binary):
                    raise CliError(f"Unable to find {binary} in your path!")
            return fn(*args, **kwargs)

        return wrapped

    return decorate


def require_login(fn: F) -> F:
    @click.pass_context
    @wraps(fn)
    def wrapped(context: click.Context, *args: Any, **kwargs: Any) -> Any:
        if not Config.is_logged_in():
            raise CliError("Please run `sym login` first")
        return context.invoke(fn, *args, **kwargs)

    return cast(F, wrapped)


def loses_interactivity(fn: F) -> F:
    @click.pass_context
    @wraps(fn)
    def wrapped(context: click.Context, *args: Any, **kwargs: Any) -> Any:
        if (resource := kwargs.get("resource")) :
            saml_client = SAMLClientFactory.create_saml_client(
                resource, context.find_object(GlobalOptions)
            )
            saml_client.ensure_session()
        return context.invoke(fn, *args, **kwargs)

    return cast(F, require_login(wrapped))


def skip_analytics(fn: F) -> F:
    @click.pass_context
    @wraps(fn)
    def wrapped(context: click.Context, *args: Any, **kwargs: Any) -> Any:
        options = context.ensure_object(GlobalOptions)
        options.disable_analytics = True
        return context.invoke(fn, *args, **kwargs)

    return cast(F, wrapped)


def setup_segment(**kwargs: Any) -> Callable[[F], F]:
    analytics.default_client = analytics.client.Client(send=is_prod_mode(), **kwargs)

    def decorator(fn: F) -> F:
        @click.pass_context
        @wraps(fn)
        def wrapped(context: click.Context, *args: Any, **kwargs: Any) -> Any:
            segment.identify(context.find_object(GlobalOptions))
            return context.invoke(fn, *args, **kwargs)

        return cast(F, wrapped)

    return decorator


def setup_sentry(**kwargs: Any) -> Callable[[F], F]:
    environment = "pypi" if is_prod_mode() and not Config.is_sym() else "development"
    sample_rate = 1.0 if is_prod_mode() else 0.0
    sentry_init(environment=environment, sample_rate=sample_rate, **kwargs)

    def decorator(fn: F) -> F:
        @click.pass_context
        @wraps(fn)
        def wrapped(context: click.Context, *args: Any, **kwargs: Any) -> Any:
            with configure_scope() as scope:
                set_scope_tag_org(scope)
                set_scope_user(scope)
                set_scope_context_os(scope)
                return context.invoke(fn, *args, **kwargs)

        return cast(F, wrapped)

    return decorator


def retry(*exceptions, count: int = 2, delay: int = None, check_ex=lambda _: True):
    def decorator(fn: F) -> F:
        @wraps(fn)
        def wrapped(*args, **kwargs):
            tries = 0
            while tries < count:
                tries += 1
                try:
                    return fn(*args, **kwargs)
                except exceptions as ex:
                    if not check_ex(ex) or tries == count:
                        raise
                    if delay:
                        time.sleep(delay)

        return cast(F, wrapped)

    return decorator



from typing import Optional

from ...saml_clients.saml_client import SAMLClient
from ..config.config import Config
from .client import RunningInstance
from .multiregion import MultiRegionEc2Client


class CachingEc2Client(MultiRegionEc2Client):
    def __init__(self, client: SAMLClient):
        super().__init__(client)

    def _load_from_config(self, alias: str) -> Optional[RunningInstance]:
        """
        In the EC2 client, an alias could be an instance ID, IP address, or DNS name.
        Look for an instance in the config with an id that matches "alias", or with
        an entry in "aliases" that matches.
        """
        for instance_id, server in Config.get_servers().items():
            try:
                region = server["region"]
                if region:
                    if alias == instance_id:
                        return RunningInstance(instance_id=alias, region=region)

                    for known_alias in server["aliases"]:
                        if alias == known_alias:
                            return RunningInstance(instance_id=instance_id, region=region)
            except KeyError:
                self.client.dprint("found invalid cache entry - region missing")

        return None

    def _store_to_config(self, alias: str, instance: RunningInstance):
        Config.record_instance(
            instance.instance_id,
            alias=alias if alias != instance.instance_id else None,
            region=instance.region,
        )

    def load_instance_by_alias(self, alias: str) -> RunningInstance:
        """
        The CachingEc2Client.load_instance_by_alias method first attempts
        to load the instance based on the local cached configuration.
        If the alias matches the instance ID, or is a member of the aliases
        list for an instance we have visited in the past, that will be returned.
        Otherwise, the instance is loaded by delegating to a MultiRegionEc2Client,
        then stored in the config before returning.
        """
        if (existing := self._load_from_config(alias)) :
            return existing

        instance = super().load_instance_by_alias(alias)
        self._store_to_config(alias, instance)
        return instance



import click

from sym.cli.saml_clients.saml_client_factory import SAMLClientFactory

from ..decorators import command_require_bins, loses_interactivity, require_login
from ..helpers.global_options import GlobalOptions
from ..helpers.options import resource_argument
from .sym import sym


@sym.command(hidden=True, short_help="New SSM Session")
@resource_argument
@click.option(
    "--target", help="target instance id", metavar="<instance-id>", required=True
)
@click.make_pass_decorator(GlobalOptions)
@command_require_bins("aws", "session-manager-plugin")
@require_login
@loses_interactivity
def ssm(options: GlobalOptions, resource: str, target: str) -> None:
    """Use approved creds for RESOURCE to start an SSM session to an EC2 instance"""
    client = SAMLClientFactory.create_saml_client(resource, options)
    client.exec("aws", "ssm", "start-session", target=target)



import functools
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Callable, ContextManager, Iterator
from uuid import UUID, uuid4

import boto3
import click
import pytest
from _pytest.monkeypatch import MonkeyPatch
from botocore.stub import Stubber
from click.testing import CliRunner

from sym.cli.helpers import boto
from sym.cli.helpers import os as sym_os
from sym.cli.helpers.config import Config
from sym.cli.helpers.envvar_option import reset_used
from sym.cli.helpers.global_options import GlobalOptions
from sym.cli.helpers.sym_group import SymGroup
from sym.cli.saml_clients.aws_okta import AwsOkta
from sym.cli.saml_clients.aws_profile import AwsProfile
from sym.cli.saml_clients.saml_client import SAMLClient
from sym.cli.sym import sym as click_command
from sym.cli.tests.helpers.capture import CaptureCommand
from sym.cli.tests.helpers.sandbox import Sandbox

CustomOrgFixture = Callable[[str], ContextManager[None]]


@pytest.fixture(autouse=True)
def patch_is_setup(monkeypatch: MonkeyPatch):
    monkeypatch.setattr(SAMLClient, "check_is_setup", lambda self: ...)


@pytest.fixture(autouse=True)
def patch_execvp(monkeypatch: MonkeyPatch):
    monkeypatch.setattr(sym_os, "execvp", lambda client, args: client.exec(*args))


@pytest.fixture
def sandbox(tmp_path: Path) -> Sandbox:
    return Sandbox(tmp_path)


@pytest.fixture
def uuid() -> UUID:
    return uuid4()


@pytest.fixture
def uuid_factory() -> Callable[[], UUID]:
    return uuid4


@pytest.fixture
def custom_org(monkeypatch: MonkeyPatch) -> CustomOrgFixture:
    @contextmanager
    def custom_org(org: str) -> Iterator[None]:
        with monkeypatch.context() as mp:
            mp.setattr(Config, "get_org", classmethod(lambda cls: org))
            yield

    return custom_org


@pytest.fixture
def capture_command(monkeypatch: MonkeyPatch) -> CaptureCommand:
    return CaptureCommand(monkeypatch)


@pytest.fixture
def click_context(sandbox):
    with sandbox.push_xdg_config_home():
        Config.instance()["org"] = "sym"
        Config.instance()["email"] = "y@symops.io"
        sandbox.create_binary(f"bin/{AwsOkta.binary}")
        with sandbox.push_exec_path():
            with click.Context(click_command) as ctx:
                ctx.ensure_object(GlobalOptions)
                yield ctx


@pytest.fixture
def wrapped_cli_runner():
    """Yield a click.testing.CliRunner to invoke the CLI."""
    class_ = CliRunner

    def invoke_wrapper(f):
        """Augment CliRunner.invoke to emit its output to stdout.

        This enables pytest to show the output in its logs on test
        failures. Otherwise, the isolated environment swallows everything.

        Seen here: https://github.com/pallets/click/issues/737

        Example:
            def command_login_tester_with_echo(click_setup):
                with click_setup() as runner:
                    result = runner.invoke(click_command, args, echo=True)
        """

        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            echo = kwargs.pop("echo", False)
            result = f(*args, **kwargs)

            if echo is True:
                sys.stdout.write(result.output)

            return result

        return wrapper

    class_.invoke = invoke_wrapper(class_.invoke)
    cli_runner = class_()

    yield cli_runner


@pytest.fixture
def click_setup(sandbox: Sandbox, wrapped_cli_runner):
    @contextmanager
    def context(set_org=True, set_client=True):
        runner = wrapped_cli_runner
        with runner.isolated_filesystem():
            with sandbox.push_xdg_config_home():
                if set_client:
                    sandbox.create_binary(f"bin/{AwsOkta.binary}")
                with sandbox.push_exec_path():
                    if set_org:
                        Config.instance()["org"] = "sym"
                        Config.instance()["email"] = "y@symops.io"
                    yield runner
        reset_used()

    return context


def _env_str(creds):
    return "\n".join([f"{k}={v}" for k, v in creds.items()]) + "\n"


@pytest.fixture
def creds_env():
    return {
        "AWS_REGION": "us-east-2",
        "AWS_ACCESS_KEY_ID": "ASIA4GNNUMIHHBPSAQC3",
        "AWS_SECRET_ACCESS_KEY": "xxx",
        "AWS_SESSION_TOKEN": "xxx",
        "AWS_OKTA_SESSION_EXPIRATION": "1600494616",
    }


@pytest.fixture
def creds_env_str(creds_env):
    return _env_str(creds_env)


@pytest.fixture
def env_creds(creds_env_str, capture_command: CaptureCommand):
    """Include this fixture to automatically stub the subprocess
    that retrieves credentials from the environment.

    Will only work when subprocess is executed with capture_command
    context. For example:

    def test_with_creds(env_creds, capture_command):
        with capture_command():
            execute thing that gets creds

    will successfully use this stub. However this will not:

    def test_with_creds(env_creds):
        execute thing that gets creds
    """
    capture_command.register_output(r"env", creds_env_str)


@pytest.fixture
def fake_creds_env():
    return {
        "AWS_REGION": "foobar",
        "AWS_FOOBAR": "baz",
        "AWS_OKTA_SESSION_EXPIRATION": "1600494616",
    }


@pytest.fixture
def fake_creds_env_str(fake_creds_env):
    return _env_str(fake_creds_env)


@pytest.fixture
def saml_client(sandbox: Sandbox, monkeypatch: MonkeyPatch, creds_env):
    def get_creds(self):
        return dict(creds_env)

    monkeypatch.setattr(AwsOkta, "get_creds", get_creds)

    def profile_matches_caller_identity(self):
        return True

    monkeypatch.setattr(
        AwsOkta, "_profile_matches_caller_identity", profile_matches_caller_identity
    )

    sandbox.create_binary(f"bin/{AwsOkta.binary}")
    with sandbox.push_exec_path():
        return AwsOkta("test", options=GlobalOptions(debug=False))


@pytest.fixture
def ssm_bins(sandbox: Sandbox):
    for binary in ["aws", "session-manager-plugin", "ssh"]:
        sandbox.create_binary(f"bin/{binary}")


@pytest.fixture
def boto_stub(monkeypatch: MonkeyPatch):
    stubs = {}

    def boto_client(_saml_client, service):
        if service in stubs:
            return stubs[service][0]

        client = boto3.client(service)
        stubber = Stubber(client)
        stubber.activate()

        stubs[service] = (client, stubber)
        return client

    monkeypatch.setattr(boto, "boto_client", boto_client)

    def get_stub(service):
        boto_client(None, service)
        return stubs[service][1]

    return get_stub


@contextmanager
def setup_context():
    with click.Context(click_command) as ctx:
        ctx.obj = GlobalOptions(saml_client_type=AwsProfile)
        yield


def empty_saml_client():
    return AwsProfile("test", options=GlobalOptions())


@pytest.fixture(autouse=True)
def cleanup(request):
    request.addfinalizer(SymGroup.reset_tees)



import pytest

from sym.cli.saml_clients.aws_profile import AwsProfile
from sym.cli.tests.helpers.capture import CaptureCommand
from sym.cli.tests.saml_clients.conftest import TestContextFixture

pytestmark = pytest.mark.usefixtures("click_context")


@pytest.mark.parametrize(argnames=["constructor"], argvalues=[[AwsProfile]])
def test_aws_profile(
    test_context_with_creds: TestContextFixture[AwsProfile],
    capture_command: CaptureCommand,
) -> None:
    with test_context_with_creds(debug=False) as aws_profile:
        with capture_command():
            aws_profile.exec("aws", "ssm", "start-session", target="i-0123456789abcdef")
    capture_command.assert_command("aws ssm start-session --target i-0123456789abcdef")


@pytest.mark.parametrize(argnames=["constructor"], argvalues=[[AwsProfile]])
def test_aws_profile_debug(
    test_context_with_creds: TestContextFixture[AwsProfile],
    capture_command: CaptureCommand,
) -> None:
    with test_context_with_creds(debug=True) as aws_profile:
        with capture_command():
            aws_profile.exec("aws", "ssm", "start-session", target="i-0123456789abcdef")
    capture_command.assert_command("aws ssm start-session --target i-0123456789abcdef")



from typing import Type

import pytest

from sym.cli.errors import SAMLClientNotFound
from sym.cli.saml_clients.aws_okta import AwsOkta
from sym.cli.saml_clients.aws_profile import AwsProfile
from sym.cli.saml_clients.chooser import SAMLClientName, choose_saml_client
from sym.cli.saml_clients.saml2aws import Saml2Aws
from sym.cli.saml_clients.saml_client import SAMLClient
from sym.cli.tests.helpers.sandbox import Sandbox


@pytest.mark.parametrize(
    argnames=("saml_client_name", "make_saml2aws", "expected"),
    argvalues=[
        ("auto", True, Saml2Aws),
        ("aws-profile", False, AwsProfile),
        ("aws-profile", True, AwsProfile),
        ("aws-okta", False, AwsOkta),
        ("aws-okta", True, AwsOkta),
        ("saml2aws", False, Saml2Aws),
        ("saml2aws", True, Saml2Aws),
    ],
)
def test_chooser_selects_available_saml_client(
    sandbox: Sandbox,
    saml_client_name: SAMLClientName,
    make_saml2aws: bool,
    expected: Type[SAMLClient],
) -> None:
    if make_saml2aws:
        sandbox.create_file("bin/saml2aws", 0o755)
    with sandbox.push_exec_path():
        assert choose_saml_client(saml_client_name) == expected


def test_chooser_auto_no_client_installed_raises_exception(sandbox: Sandbox):
    with pytest.raises(SAMLClientNotFound):
        with sandbox.push_exec_path():
            choose_saml_client("auto")



import datetime
from typing import Optional, Sequence

import pytest

from sym.cli.helpers.config import Config
from sym.cli.helpers.ec2.cache import CachingEc2Client
from sym.cli.saml_clients.saml_client import SAMLClient
from sym.cli.tests.helpers.ec2.conftest import (
    BOTO_INSTANCE_RESPONSE,
    DEFAULT_REGION,
    TEST_INSTANCE,
    TEST_INSTANCE_ID,
    TEST_IP,
)
from sym.cli.tests.helpers.sandbox import Sandbox


@pytest.fixture
def caching_ec2_client(saml_client: SAMLClient) -> CachingEc2Client:
    return CachingEc2Client(saml_client)


def assert_no_cached_instances():
    cached = Config.get_servers()
    assert len(cached) == 0


def assert_cache_miss(alias: str, caching_ec2_client: CachingEc2Client, ec2_stub):
    assert_no_cached_instances()
    ec2_stub.add_response(
        "describe_instances",
        BOTO_INSTANCE_RESPONSE,
    )
    instance = caching_ec2_client.load_instance_by_alias(alias)
    assert instance == TEST_INSTANCE
    ec2_stub.assert_no_pending_responses()


def assert_cache_hit(caching_ec2_client: CachingEc2Client):
    instance = caching_ec2_client.load_instance_by_alias(TEST_INSTANCE_ID)
    assert instance == TEST_INSTANCE


def assert_cached_instance(
    aliases: Sequence[str] = [],
    last_connection: Optional[datetime.datetime] = None,
    total_instances: int = 1,
):
    cached = Config.get_servers()
    assert len(cached) == total_instances

    cached_instance = cached[TEST_INSTANCE_ID]
    assert cached_instance["region"] == DEFAULT_REGION
    assert cached_instance["aliases"] == aliases
    assert cached_instance["last_connection"] == last_connection


def test_cache_miss_then_hit_id(
    sandbox: Sandbox, caching_ec2_client: CachingEc2Client, ec2_stub
):
    with sandbox.push_xdg_config_home():
        assert_cache_miss(TEST_INSTANCE_ID, caching_ec2_client, ec2_stub)
        assert_cached_instance()
        assert_cache_hit(caching_ec2_client)


def test_cache_miss_then_hit_by_ip(
    sandbox: Sandbox, caching_ec2_client: CachingEc2Client, ec2_stub
):
    with sandbox.push_xdg_config_home():
        assert_cache_miss(TEST_IP, caching_ec2_client, ec2_stub)
        assert_cached_instance(aliases=[TEST_IP])
        assert_cache_hit(caching_ec2_client)


TEST_MISSING_REGION_CACHED = f"""
email: r@symops.io
org: sym
servers:
  {TEST_INSTANCE_ID}:
    aliases: 
      - {TEST_IP}
    last_connection: 2020-11-12 10:16:47.179202
"""


def test_cache_hit_with_missing_region(
    sandbox: Sandbox, caching_ec2_client: CachingEc2Client, ec2_stub
):
    with sandbox.push_xdg_config_home():
        with Config.instance().file as f:
            f.write(TEST_MISSING_REGION_CACHED)

        assert_cache_miss(TEST_INSTANCE_ID, caching_ec2_client, ec2_stub)
        assert_cached_instance(
            aliases=[TEST_IP],
            last_connection=datetime.datetime(2020, 11, 12, 10, 16, 47, 179202),
        )
        assert_cache_hit(caching_ec2_client)


TEST_NULL_REGION_CACHED = f"""
email: r@symops.io
org: sym
servers:
  {TEST_INSTANCE_ID}:
    aliases: 
      - {TEST_IP}
    last_connection: 2020-11-12 10:16:47.179202
    region: null
"""


def test_cache_hit_with_null_region(
    sandbox: Sandbox, caching_ec2_client: CachingEc2Client, ec2_stub
):
    with sandbox.push_xdg_config_home():
        with Config.instance().file as f:
            f.write(TEST_NULL_REGION_CACHED)

        assert_cache_miss(TEST_INSTANCE_ID, caching_ec2_client, ec2_stub)
        assert_cached_instance(
            aliases=[TEST_IP],
            last_connection=datetime.datetime(2020, 11, 12, 10, 16, 47, 179202),
        )
        assert_cache_hit(caching_ec2_client)



import json

from sym.cli.helpers.ec2.factory import get_ec2_client
from sym.cli.saml_clients.saml_client_factory import SAMLClientFactory

from ...errors import BotoError, CliError, InstanceNotFound
from ..boto import get_identity
from .model import CheckContext, CheckResult, SymCheck, failure, success


class CallerIdentityCheck(SymCheck):
    def check(self, ctx: CheckContext) -> CheckResult:
        client = SAMLClientFactory.create_saml_client(ctx.resource, ctx.options)
        try:
            identity = get_identity(client)
            pretty = json.dumps(identity, indent=4, sort_keys=True)
            return success(f"Successfully got AWS identity:\n{pretty}")
        except CliError as e:
            return failure(f"Unable to get AWS identity: {e.message}")


class DescribeRegionsCheck(SymCheck):
    def check(self, ctx: CheckContext) -> CheckResult:
        client = SAMLClientFactory.create_saml_client(ctx.resource, ctx.options)

        try:
            ctx.regions = get_ec2_client(client).get_regions()
            return success(f"Found {len(ctx.regions)} regions")
        except BotoError as e:
            return failure(f"Unable to describe regions: {e.message}")


class DescribeInstanceCheck(SymCheck):
    def __init__(self, instance: str):
        self.instance = instance

    def check(self, ctx: CheckContext) -> CheckResult:
        client = SAMLClientFactory.create_saml_client(ctx.resource, ctx.options)
        try:
            running_instance = get_ec2_client(client).load_instance_by_alias(
                self.instance
            )
            return success(
                f"Located instance {running_instance.instance_id} in region {running_instance.region}"
            )
        except InstanceNotFound as e:
            return failure(f"{e.message}")



from typing import Dict, NamedTuple, Optional, TypedDict

import validators

from ..errors import CliError
from .config import Config


class Profile(NamedTuple):
    display_name: str
    region: str
    arn: str
    aws_saml_url: Optional[str] = None
    ansible_bucket: Optional[str] = None


class OrgParams(TypedDict, total=False):
    resource_env_var: Optional[str]
    users: Dict[str, str]
    domain: str
    aws_saml_url: str
    aws_okta_params: Dict[str, str]
    saml2aws_params: Dict[str, str]
    profiles: Dict[str, Profile]


# For legacy reasons, we need to maintain multiple org slugs for LD.
# This allows us to put the config in one place.
LaunchDarklyParams: OrgParams = {
    "resource_env_var": "ENVIRONMENT",
    "users": {"ssh": "ubuntu", "ansible": "ansible"},
    "domain": "launchdarkly.com",
    "aws_saml_url": (
        "https://launchdarkly.okta.com/home/amazon_aws/0oaj4aow7gPk26Fy6356/272"
    ),
    "aws_okta_params": {
        "mfa_provider": "OKTA",
        "mfa_factor_type": "push",
        "assume_role_ttl": "1h",
        "session_ttl": "30m",
    },
    "saml2aws_params": {"mfa": "PUSH", "aws_session_duration": "1800"},
    "profiles": {
        "production": Profile(
            display_name="Production",
            region="us-east-1",
            arn="arn:aws:iam::554582317989:role/ops/SSHAdmin-production",
            ansible_bucket="launchdarkly-sym-ansible-production",
        ),
        "staging": Profile(
            display_name="Staging",
            region="us-east-1",
            arn="arn:aws:iam::554582317989:role/ops/SSHAdmin-staging",
            ansible_bucket="launchdarkly-sym-ansible-staging",
        ),
        "catamorphic": Profile(
            display_name="Catamorphic",
            region="us-east-1",
            arn="arn:aws:iam::554582317989:role/ops/SSHAdmin-catamorphic",
            ansible_bucket="launchdarkly-sym-ansible-catamorphic",
        ),
        "prod-intuit": Profile(
            display_name="Prod: Intuit",
            region="us-west-2",
            arn="arn:aws:iam::527291094460:role/ops/SSHAdmin-production",
            ansible_bucket="launchdarkly-sym-ansible-prod-intuit",
        ),
        "staging-intuit": Profile(
            display_name="Staging: Intuit",
            region="us-west-2",
            arn="arn:aws:iam::527291094460:role/ops/SSHAdmin-staging",
            ansible_bucket="launchdarkly-sym-ansible-staging-intuit",
        ),
        "production_apac": Profile(
            display_name="Production: APAC",
            region="ap-southeast-1",
            arn="arn:aws:iam::554582317989:role/ops/SSHAdmin-production",
            ansible_bucket="launchdarkly-sym-ansible-production",
        ),
        "production_euw1": Profile(
            display_name="Production: EU West 1",
            region="eu-west-1",
            arn="arn:aws:iam::554582317989:role/ops/SSHAdmin-production",
            ansible_bucket="launchdarkly-sym-ansible-production",
        ),
        "staging_apac": Profile(
            display_name="Staging: APAC",
            region="ap-southeast-1",
            arn="arn:aws:iam::554582317989:role/ops/SSHAdmin-staging",
            ansible_bucket="launchdarkly-sym-ansible-staging",
        ),
        "staging_euw1": Profile(
            display_name="Staging: EU West 1",
            region="eu-west-1",
            arn="arn:aws:iam::554582317989:role/ops/SSHAdmin-staging",
            ansible_bucket="launchdarkly-sym-ansible-staging",
        ),
    },
}

PARAMS: Dict[str, OrgParams] = {
    "launch-darkly": LaunchDarklyParams,
    "launchdarkly": LaunchDarklyParams,
    "sym": {
        "resource_env_var": "ENVIRONMENT",
        "users": {"ssh": "ubuntu", "ansible": "ubuntu"},
        "domain": "symops.io",
        "aws_saml_url": (
            "https://dev-291131.okta.com/home/amazon_aws/0oaqlmsn7GMVgAyBK4x6/272"
        ),
        "aws_okta_params": {
            "mfa_provider": "OKTA",
            "mfa_factor_type": "push",
            "assume_role_ttl": "1h",
            "session_ttl": "30m",
        },
        "saml2aws_params": {"mfa": "PUSH", "aws_session_duration": "1800"},
        "profiles": {
            "test": Profile(
                display_name="Test",
                region="us-east-1",
                arn="arn:aws:iam::838419636750:role/SSMTestRole",
            ),
            "test_euw1": Profile(
                display_name="Test: EU West 1",
                region="eu-west-1",
                arn="arn:aws:iam::838419636750:role/SSMTestRole",
            ),
            "test-custom": Profile(
                display_name="TestCustom",
                region="us-east-1",
                arn="arn:aws:iam::838419636750:role/SSMTestRoleCustomBucket",
                ansible_bucket="sym-ansible-dev",
            ),
        },
    },
    "healthy-health": {
        "resource_env_var": "ENVIRONMENT",
        "users": {"ssh": "ubuntu"},
        "domain": "healthy-health.co",
        "aws_saml_url": (
            "https://healthy-health.okta.com/home/amazon_aws/0oagrj7yFaVtJoI2N5d5/272"
        ),
        "aws_okta_params": {
            "mfa_provider": "OKTA",
            "mfa_factor_type": "push",
            "assume_role_ttl": "1h",
            "session_ttl": "30m",
        },
        "saml2aws_params": {"mfa": "PUSH", "aws_session_duration": "1800"},
        "profiles": {
            "prod": Profile(
                display_name="Prod",
                region="us-east-1",
                arn="arn:aws:iam::223440862848:role/ssm-healthy-health-user-prod",
            ),
        },
    },
}


def get_org_params() -> OrgParams:
    return PARAMS[Config.get_org()]


def get_aws_saml_url(resource: str) -> str:
    org_params = get_org_params()
    profile = org_params["profiles"][resource]
    return profile.aws_saml_url or org_params["aws_saml_url"]


def get_aws_okta_params() -> Dict[str, str]:
    return get_org_params()["aws_okta_params"]


def get_saml2aws_params() -> Dict[str, str]:
    return get_org_params()["saml2aws_params"]


def get_profiles() -> Dict[str, Profile]:
    return get_org_params()["profiles"]


def get_domain() -> str:
    return get_org_params()["domain"]


def get_ssh_user() -> str:
    return get_org_params()["users"]["ssh"]


def get_ansible_user() -> str:
    return get_org_params()["users"]["ansible"]


def get_resource_env_vars() -> Optional[str]:
    try:
        return ("SYM_RESOURCE", get_org_params()["resource_env_var"])
    except KeyError:
        return "SYM_RESOURCE"


def get_profile(resource: str) -> Profile:
    return get_profiles()[resource]


def set_login_fields(org: str, email: str):
    if org not in PARAMS:
        raise CliError(f"Invalid org: {org}")

    config = Config.instance()
    config["org"] = org

    if not validators.email(email):
        raise CliError(f"Invalid email: {email}")

    if (domain := email.split("@")[-1]) != get_domain():
        raise CliError(f"Invalid domain: {domain}")

    config["email"] = email



from typing import List

from sym.cli.actions.sym_action import SymAction
from sym.cli.data.request_data import RequestData
from sym.cli.helpers.check.core import DependenciesCheck, SymCheck
from sym.cli.helpers.ec2.factory import get_ec2_client
from sym.cli.helpers.ssh import start_tunnel
from sym.cli.saml_clients.saml_client_factory import SAMLClientFactory


class SSHSessionAction(SymAction):
    def __init__(self):
        super().__init__("ssh_session")

    def get_requirements(self) -> List[SymCheck]:
        return [DependenciesCheck("aws"), DependenciesCheck("session-manager-plugin")]

    def execute(self, request_data: RequestData):
        """Use approved creds for RESOURCE to tunnel a SSH session through an SSM session."""
        client = SAMLClientFactory.create_saml_client(
            request_data.resource, request_data.get_global_options()
        )

        target_options = request_data.get_target_options()
        running_instance = get_ec2_client(client).load_instance_by_alias(
            target_options.host
        )

        new_client = SAMLClientFactory.clone(client, aws_region=running_instance.region)
        start_tunnel(new_client, running_instance.instance_id, target_options.port)



from threading import Condition, Thread
from typing import Callable, Optional, Sequence, Tuple

import click

from sym.cli.errors import CliError


def first(
    values: Sequence[str],
    finder: Callable[[str], str],
    timeout=5,
) -> Optional[Tuple[str, str]]:
    ctx = click.get_current_context()
    results = {}
    cv = Condition()

    def target(val):
        with ctx:
            try:
                res = finder(val)
            except CliError:
                return
            if res:
                with cv:
                    results[val] = res
                    cv.notify()

    for val in values:
        Thread(target=target, args=(val,), daemon=True).start()

    with cv:
        if cv.wait_for(lambda: bool(results), timeout=timeout):
            return list(results.items())[0]



import json
from multiprocessing import Pool
from uuid import UUID

import pytest

from sym.cli.helpers.config import Config, SymConfigFile, init
from sym.cli.helpers.contexts import push_env
from sym.cli.tests.helpers.sandbox import Sandbox


@pytest.fixture
def org(sandbox: Sandbox, uuid: UUID) -> str:
    org = uuid.hex
    with sandbox.create_file_with_content(".config/sym/default/config.yml") as f:
        print(f"org: {org}", file=f)
    return org


def test_get_org_with_xdg_config_home(sandbox: Sandbox, org: str) -> None:
    with push_env("XDG_CONFIG_HOME", str(sandbox.path / ".config")):
        assert Config()["org"] == org


def test_get_org_without_xdg_config_home(sandbox: Sandbox, org: str) -> None:
    with push_env("HOME", str(sandbox.path)):
        assert Config()["org"] == org


def test_org_is_sym(sandbox: Sandbox):
    with sandbox.push_xdg_config_home():
        Config.instance()["org"] = "sym"
        assert Config.is_sym() is True


def test_org_is_not_sym(sandbox: Sandbox):
    with sandbox.push_xdg_config_home():
        Config.instance()["org"] = "notsym"
        assert Config.is_sym() is False


def test_sym_config_truncate(sandbox: Sandbox):
    short_json = json.dumps({"hello": "world"})
    long_json = json.dumps({"hello": "world", "foo": "bar"})

    with sandbox.push_xdg_config_home():
        config_file = SymConfigFile(a=1, b=2, file_name="config.json")
        with config_file as f:
            f.write(long_json)
            f.seek(0)
            assert f.read() == long_json
        with config_file as f:
            assert f.read() == long_json
        with config_file as f:
            f.write(short_json)
            f.seek(0)
            assert f.read() == short_json
        with config_file as f:
            assert f.read() == short_json


def _exclusive_create_write_fn(args):
    file_name, body = args

    init("sym")  # We're not sure why this is needed, something about ClassVar + Pool
    config_file = SymConfigFile(file_name=file_name)
    with config_file.exclusive_create() as f:
        f.write(body)


def test_sym_config_exclusive_create_already_created(sandbox: Sandbox):
    file_name = "config.json"
    body = json.dumps({"hello": "world"})

    with sandbox.push_xdg_config_home():
        with Pool(processes=4) as pool:
            pool.map(_exclusive_create_write_fn, [(file_name, body)] * 100)

        with SymConfigFile(file_name=file_name) as f:
            assert f.read() == body



from contextlib import contextmanager
from typing import List, Sequence

import pytest
from click.testing import Result as ClickTestResult
from expects import *

from sym.cli.helpers.contexts import push_envs
from sym.cli.saml_clients.saml_client import SAMLClient
from sym.cli.sym import sym as click_command
from sym.cli.tests.helpers.capture import CaptureCommand
from sym.cli.tests.helpers.sandbox import Sandbox
from sym.cli.tests.matchers import fail_with, succeed


@pytest.fixture
def command_login_tester(click_setup):
    def tester(args, envs={}):
        with click_setup(set_org=False) as runner:
            with push_envs(envs):
                result = runner.invoke(click_command, args)
                assert result.exit_code > 0
                assert "Error: Please run `sym login` first" in result.output

    return tester


@pytest.fixture
def command_tester(
    click_context,
    click_setup,
    ssm_bins,
    env_creds,
    boto_stub,
    saml_client: SAMLClient,
    capture_command: CaptureCommand,
    sandbox: Sandbox,
):
    def tester(command: Sequence[str], setup=None, **kwargs):
        stubs = []

        def make_stub(service):
            stub = boto_stub(service)
            stubs.append(stub)
            return stub

        if setup:
            setup()

        teardown_ = kwargs.get("teardown")

        @contextmanager
        def context(setup=None, teardown=None, exception=None):
            with click_setup() as runner:
                with click_context:
                    if setup:
                        setup(make_stub)

                with capture_command():
                    result = runner.invoke(
                        click_command, command, catch_exceptions=False, echo=True
                    )
                    if exception:
                        expect(result).to(fail_with(exception))
                    else:
                        expect(result).to(succeed())
                    yield result

                for stub in stubs:
                    stub.assert_no_pending_responses()

                if teardown:
                    teardown()
                if teardown_:
                    teardown_()

        return context

    return tester


@pytest.fixture
def simple_command_tester(wrapped_cli_runner):
    """Run the click CliRunner with a custom invoke method.

    Use this fixture rather than command_tester to allow other fixtures
    access to information about processes and functions inside the CliRunner
    isolated environment. For example, if you need to know what subprocesses
    were run, or what functions were called via patch.

    Example:
        # in this version, assert_called_once() will succeed
        # mock_execute_write_creds.call_count == 1
        with patch.object(CommandHandler, "execute_write_creds") as mock_execute_write_creds:
            result = simple_command_tester([
                "write-creds",
                "test",
                "--profile",
                "test-profile",
                "--path",
                 "test/path"
            ])
            mock_execute_write_creds.assert_called_once()

        # in this version, assert_called_once() will fail
        # mock_execute_write_creds.call_count == 0
        with patch.object(CommandHandler, "execute_write_creds") as mock_execute_write_creds:
            command_tester(["write-creds", "test", "--profile", "test-profile", "--path", "test/path"])
            mock_execute_write_creds.assert_called_once()
    """

    def f(
        click_args: List[str],
    ) -> ClickTestResult:
        """
        Args:
            echo: Allow prints from .invoke to be shown
        """

        return wrapped_cli_runner.invoke(
            click_command, click_args, catch_exceptions=False
        )

    return f



from typing import Literal, Optional, Type

from ..errors import SAMLClientNotFound
from ..helpers.os import has_command
from ..helpers.util import requires_all_imports
from . import import_all
from .saml_client import SAMLClient

SAMLClientName = Literal["auto", "aws-okta", "saml2aws", "aws-profile"]
AUTO_EXCLUDED_SAML_CLIENT_NAMES = ["aws-profile"]


@requires_all_imports(import_all)
def option_values():
    return ["auto"] + [x.option_value for x in SAMLClient.sorted_subclasses()]


@requires_all_imports(import_all)
def choose_saml_client(
    saml_client_name: SAMLClientName, none_ok=False
) -> Optional[Type[SAMLClient]]:
    if saml_client_name == "auto":
        for client in SAMLClient.sorted_subclasses():
            if (
                client.option_value not in AUTO_EXCLUDED_SAML_CLIENT_NAMES
                and has_command(client.binary)
            ):
                return client
    else:
        for client in SAMLClient.sorted_subclasses():
            if client.option_value == saml_client_name:
                return client

    if not none_ok:
        raise SAMLClientNotFound()



import os
import re
import subprocess
import sys
from datetime import timedelta
from pathlib import Path
from signal import SIGINT, SIGPIPE, signal
from subprocess import CalledProcessError
from textwrap import dedent
from typing import List, Sequence, Tuple

import click

from sym.cli.helpers.contexts import push_env
from sym.cli.helpers.global_options import GlobalOptions

from ..decorators import intercept_errors, retry, run_subprocess
from ..errors import (
    AccessDenied,
    FailedSubprocessError,
    MissingPublicKey,
    SuppressedError,
    TargetNotConnected,
    WrappedSubprocessError,
)
from ..helpers.boto import send_ssh_key, ssm_interactive_command, ssm_ssh_session
from ..helpers.config import Config, SymConfigFile
from ..helpers.params import get_ssh_user
from ..helpers.sym_group import SymGroup
from ..saml_clients.saml_client import SAMLClient

MissingPublicKeyPattern = re.compile(r"Permission denied \(.*publickey.*\)")
TargetNotConnectedPattern = re.compile("TargetNotConnected")
AccessDeniedPattern = re.compile("AccessDeniedException")
ConnectionClosedPattern = re.compile(r"Connection to .* closed")

SSHConfigPath = "ssh/config"
SSHKeyPath = "ssh/key"

OPENSSH_VERSION_PATTERN = re.compile(r"OpenSSH_(\d)\.(\d)")
MIN_VER_FOR_CONTROL_MASTER = (8, 1)


def check_ssh_version() -> bool:
    ssh_ok = get_ssh_version() >= MIN_VER_FOR_CONTROL_MASTER
    if not ssh_ok:
        click.secho(
            "OpenSSH version is out of date, control master will be disabled.",
            fg="yellow",
        )
        version = ".".join(map(str, MIN_VER_FOR_CONTROL_MASTER))
        click.secho(
            f"This will negatively affect performance, please consider upgrading to OpenSSH {version}+.",
            fg="yellow",
        )
    return ssh_ok


def get_ssh_version() -> Tuple[int, int]:
    result = subprocess.run(["ssh", "-V"], text=True, capture_output=True).stderr
    if (match := OPENSSH_VERSION_PATTERN.search(result)) :
        return tuple(map(int, (match[1], match[2])))
    return (0, 0)


def ssh_key_and_config(client: SAMLClient):
    ssh_key = SymConfigFile(file_name=SSHKeyPath, uid_scope=False)
    ssh_config = client.subconfig(SSHConfigPath, ssh_key=str(ssh_key))
    return (ssh_key, ssh_config)


@intercept_errors()
@run_subprocess
def _gen_ssh_key(dest: SymConfigFile):
    with dest.exclusive_create() as f:
        Path(f.name).unlink(missing_ok=True)
        yield "ssh-keygen", {"t": "rsa", "f": f.name, "N": ""}


def gen_ssh_key(dest: SymConfigFile, options: GlobalOptions):
    try:
        _gen_ssh_key(
            dest,
            capture_output_=True,
            input_="n\n",
            run_subprocess_options_=options,
            intercept_errors_options_=options,
        )
    except FailedSubprocessError:
        if not dest.path.exists():
            raise


def maybe_gen_ssh_key(client):
    ssh_key, _ = ssh_key_and_config(client)
    if not ssh_key.path.exists():
        gen_ssh_key(ssh_key, client.options)


def ssh_args(client, instance, port) -> tuple:
    _, ssh_config = ssh_key_and_config(client)
    opts = {
        "p": str(port),
        "F": str(ssh_config),
        "l": get_ssh_user(),
        "v": client.debug,
    }
    return (
        "ssh",
        instance,
        opts,
        f"-o HostKeyAlias={instance}",
    )


@run_subprocess
def _start_background_ssh_session(client: SAMLClient, instance: str, port: int, *command):
    with push_env("SHELL", "/bin/bash"):
        yield (
            *ssh_args(client, instance, port),
            {"f": True},
            "-o BatchMode=yes",
            *command,
        )


@run_subprocess
def _start_ssh_session(client: SAMLClient, instance: str, port: int, *command: str):
    with push_env("SHELL", "/bin/bash"):
        yield (*ssh_args(client, instance, port), *command)


@intercept_errors(suppress=True)
@run_subprocess
def raw_ssh(*args):
    yield ("ssh", *args)


def _parse_options_from_args(args: Sequence[str]):
    opts = {}
    for i, arg in enumerate(args):
        if arg == "-o" and i + 1 < len(args):
            k, v = re.split(r"=|\s+", args[i + 1])
            opts[k.casefold()] = (v, i)
    return opts


def _on_ssh_error(client: SAMLClient, instance: str, args: Sequence[str]):
    """
    When there's an SSH error, we'll update our config to indicate we have an error in the
    instance and we'll attempt to clear out any existing control path files.
    """
    Config.touch_instance(instance, error=True)
    opts = _parse_options_from_args(args)
    if (vi := opts.get("controlpath")) :
        try:
            os.unlink(vi[0])
        except OSError as err:
            client.dprint(f"Error cleaning up controlpath: {err}")


def _preprocess_args(client: SAMLClient, instance: str, args: Sequence[str]) -> List[str]:
    """
    Checks the instance to make sure there is an existing SSH key that was used within the TTL. If not,
    delete the ControlMaster and ControlPath arguments. Attempting to use an invalid ControlPath may
    cause the SSH client to hang.
    """
    processed = list(args)
    opts = _parse_options_from_args(args)
    if not Config.check_instance_ttl(instance):
        for opt in ("controlmaster", "controlpath"):
            if (vi := opts.get(opt)) :
                del processed[vi[1] : vi[1] + 2]

        processed.extend(["-o", "ControlMaster=no"])
        client.dprint("removed controlpath from ssh args")
    return processed


@retry(TargetNotConnected, delay=1, count=2)
@retry(MissingPublicKey, delay=0, count=2)
def start_ssh_session(
    client: SAMLClient,
    instance: str,
    port: int,
    *,
    args: Sequence[str] = [],
    command: Sequence[str] = [],
    wrap: bool = True,
):
    args = _preprocess_args(client, instance=instance, args=args)
    ensure_ssh_key(client, instance, port)
    client.dprint("starting SSH session")
    try:
        _start_ssh_session(client, instance, port, *args, *command)
    except CalledProcessError as err:
        if ConnectionClosedPattern.search(err.stderr):
            raise SuppressedError(err, echo=True) from err
        client.dprint(f"SSH Session Error: {err.stderr}")
        _on_ssh_error(client, instance, args=args)
        if MissingPublicKeyPattern.search(err.stderr):
            raise MissingPublicKey(err, get_ssh_user()) from err
        # If the ssh key path is cached then this doesn't get intercepted in ensure_ssh_key
        elif TargetNotConnectedPattern.search(err.stderr):
            raise TargetNotConnected() from err
        elif AccessDeniedPattern.search(err.stderr):
            raise AccessDenied() from err
        else:
            if wrap:
                raise WrappedSubprocessError(
                    err, f"Contact your Sym administrator.", report=True
                ) from err
            else:
                raise SuppressedError(err, echo=True)
    else:
        Config.touch_instance(instance)


def gen_local_ssh_config(client) -> SymConfigFile:
    client.dprint("writing SSH key")
    ssh_key, ssh_config = ssh_key_and_config(client)
    sym_cmd = f"PYTHONUNBUFFERED=1 sym {client.cli_options} ssh-session {client.resource} --instance %h --port %p"

    if client.debug:
        mux_config = f"""
            ControlMaster no
        """
    else:
        mux_config = f"""
            ControlMaster auto
        """

    # fmt: off
    ssh_config.put(dedent(  # Ensure the SSH Config first, always
        f"""
        Host *
            IdentityFile {str(ssh_key)}
            IdentitiesOnly yes
            PreferredAuthentications publickey
            PubkeyAuthentication yes
            StrictHostKeyChecking no
            PasswordAuthentication no
            ChallengeResponseAuthentication no
            GSSAPIAuthentication no
            ProxyCommand sh -c "{sym_cmd} || exit \\$?"
            {mux_config.strip()}
        """
    ))
    # fmt: on
    return ssh_config


def ensure_public_key_on_instance(
    client,
    instance: str,
    port: int,
    *,
    args: Sequence[str] = [],
):
    try:
        _start_background_ssh_session(
            client, instance, port, *args, "exit", capture_output_=True
        )
    except CalledProcessError as err:
        client.dprint(f"SSH Check Error: {err.stdout}")
        if not MissingPublicKeyPattern.search(err.stderr):
            raise
        ssh_key, _ = ssh_key_and_config(client)
        send_ssh_key(client, instance, ssh_key)
    else:
        client.dprint("remote SSH key check succeeded")
        Config.touch_instance(instance)


@retry(TargetNotConnected, delay=1, count=2)
@intercept_errors({TargetNotConnectedPattern: TargetNotConnected}, suppress=True)
def ensure_ssh_key(
    client,
    instance: str,
    port: int,
    *,
    ttl: timedelta = timedelta(days=1),
    args: Sequence[str] = [],
):
    gen_local_ssh_config(client)
    maybe_gen_ssh_key(client)

    if Config.check_instance_ttl(instance, ttl):
        client.dprint(f"Skipping remote SSH key check for {instance}")
        return

    ensure_public_key_on_instance(client, instance, port, args=args)


def _set_handlers(instance):
    def pipe_handler(signum, frame):
        devnull = os.open(os.devnull, os.O_WRONLY)
        os.dup2(devnull, sys.stdout.fileno())
        # Set error state on session end.
        # If the connection was actually successful,
        # the parent thread will re-set a success state.
        # This is the only reliable way to avoid error
        # loops with Ansible due to missing SSH public keys.
        Config.touch_instance(instance, error=True)

    def int_handler(signum, frame):
        # Ignore Control-Cs.
        # Let them be handled by the remote shell
        pass

    signal(SIGPIPE, pipe_handler)
    signal(SIGINT, int_handler)


def start_tunnel(client, instance: str, port: int):
    client.dprint("execing aws ssm")
    SymGroup.reset_tees()
    _set_handlers(instance)

    with ssm_ssh_session(client, instance, port) as command:
        client.exec(*command, suppress_=True)


def start_interactive_command(client, instance: str, command: Sequence[str]):
    client.dprint("execing aws ssm interactive command")
    SymGroup.reset_tees()
    _set_handlers(instance)

    with ssm_interactive_command(client, instance, get_ssh_user(), command) as command:
        client.exec(*command, suppress_=True)



import os
import shlex
import sys
from datetime import datetime
from pathlib import Path
from typing import Literal

# http://web.archive.org/web/20141016185743/https://mail.python.org/pipermail/python-list/2007-May/460639.html

TeeFD = Literal["stdout", "stderr"]


class Tee(object):
    def __init__(self, fd_name: TeeFD, log_dir: str, mode="w"):
        self.path = self.__class__.path_for_fd(log_dir, fd_name)
        self.path.parent.mkdir(parents=True, exist_ok=True)

        self.file = self.path.open(mode)
        self.fd_name = fd_name
        self.fd = getattr(sys, fd_name)

        setattr(sys, fd_name, self)

    @classmethod
    def path_for_fd(cls, log_dir, fd_name):
        return (Path(log_dir) / f"{os.getpid()}.{fd_name}.log").expanduser().resolve()

    @classmethod
    def tee_command(cls, log_dir, command):
        if isinstance(command, list):
            program = command[0]
            command = shlex.join(command)
        else:
            program = shlex.split(command)[0]
        stdout_path = cls.path_for_fd(log_dir, f"{program}.stdout")
        stderr_path = cls.path_for_fd(log_dir, f"{program}.stderr")
        return f"{command} > >(tee -a '{stdout_path}') 2> >(tee -a '{stderr_path}' >&2)"

    def close(self):
        if self.fd is not None:
            setattr(sys, self.fd_name, self.fd)
            self.fd = None

        if self.file is not None:
            if self.file.tell() == 0:
                self.path.unlink(missing_ok=True)
            else:
                self.file.write(
                    "\n".join(
                        [
                            "--",
                            f"Sym log file from {datetime.utcnow()}",
                            shlex.join(sys.argv or []),
                        ]
                    )
                )
                self.file.flush()
            self.file.close()
            self.file = None

    def write(self, data):
        self.file.write(data)
        self.fd.write(data)

    def flush(self):
        self.file.flush()
        self.fd.flush()

    def fileno(self):
        return self.fd.fileno()

    def __enter__(self):
        self.ref = self
        return self

    def __exit__(self, type, value, traceback):
        self.ref = None
        self.close()

    def __del__(self):
        self.close()


class TeeStdOut(Tee):
    def __init__(self, log_dir):
        super().__init__("stdout", log_dir)


class TeeStdErr(Tee):
    def __init__(self, log_dir):
        super().__init__("stderr", log_dir)



import click

from ..helpers.options import config_option
from ..helpers.org import infer_org_from_email
from ..helpers.params import set_login_fields
from .sym import sym


@sym.command(short_help="Log in to your sym account")
@config_option(
    "email",
    help="The email you use to log into your identity provider (e.g. Okta or GSuite).",
)
@config_option(
    "org",
    help="Your organizations's slug. Contact support if you need this.",
    default=lambda: infer_org_from_email(click.get_current_context().params["email"]),
)
def login(**kwargs) -> None:
    """Link your Sym account."""
    set_login_fields(**kwargs)
    click.echo("Sym successfully initalized!")



import json
import os
from contextlib import ExitStack
from pathlib import Path
from typing import Iterator
from unittest.mock import MagicMock, patch

import pytest
import responses
from freezegun import freeze_time

from sym.cli.helpers.updater import SymUpdater


@pytest.fixture
def time_travel() -> Iterator:
    with patch("yaml.safe_dump"):  # YAML can't serialize FakeDatetimes
        with freeze_time() as time:
            yield time


@pytest.fixture
def pypi_response() -> dict:
    return json.load((Path(__file__).parent / "response.json").open())


@pytest.fixture
def http_mock(pypi_response) -> Iterator:
    with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
        if os.getenv("CI"):
            rsps.add_passthru("http")  # make real requests on CI
        else:
            rsps.add(
                responses.GET,
                "https://pypi.org/pypi/sym-cli/json",
                json=pypi_response,
            )
        yield rsps


@pytest.fixture
def updater(sandbox, http_mock) -> Iterator[SymUpdater]:
    with ExitStack() as stack:
        stack.enter_context(sandbox.push_home())
        stack.enter_context(sandbox.push_xdg_config_home())
        stack.enter_context(sandbox.push_exec_path())
        yield MagicMock(wraps=SymUpdater)()


@pytest.fixture
def other_updater(sandbox, updater) -> SymUpdater:
    updater.dir = sandbox.create_dir("my-sym-cli")
    return updater


@pytest.fixture
def pipx_updater(sandbox, updater) -> SymUpdater:
    sandbox.create_binary("bin/pipx")
    updater.is_local = MagicMock(return_value=False)
    updater.dir = (
        sandbox.path
        / "home"
        / ".local"
        / "pipx"
        / "venvs"
        / "sym-cli"
        / "lib"
        / "python3.8"
        / "site-packages"
    )
    return updater


@pytest.fixture
def brew_updater(sandbox, updater) -> SymUpdater:
    sandbox.create_binary("bin/brew")
    updater.is_local = MagicMock(return_value=False)
    updater.dir = (
        Path("/")
        / "usr"
        / "local"
        / "Cellar"
        / "sym-cli"
        / "0.0.0"
        / "libexec"
        / "lib"
        / "python3.8"
        / "site-packages"
    )
    return updater



from sym.cli.actions.sym_action import SymAction
from sym.cli.data.request_data import RequestData
from sym.cli.helpers.ansible import run_ansible
from sym.cli.helpers.check.core import DependenciesCheck, SymCheck
from sym.cli.saml_clients.saml_client_factory import SAMLClientFactory


class AnsiblePlaybookAction(SymAction):
    def __init__(self):
        super().__init__("ansible_playbook")

    def get_requirements(self) -> [SymCheck]:
        return [
            DependenciesCheck("aws"),
            DependenciesCheck("session-manager-plugin"),
            DependenciesCheck("ansible-playbook"),
        ]

    def execute(self, request_data: RequestData) -> None:
        """Extract specifications from request_data and trigger execution
        to run an Ansible command against an inventory of EC2 instances.
        """

        global_options = request_data.get_global_options()
        ansible_options = request_data.get_ansible_options()

        client = SAMLClientFactory.create_saml_client(
            request_data.resource, global_options
        )

        run_ansible(
            client,
            ansible_options.command,
            binary="ansible-playbook",
            ansible_aws_profile=ansible_options.ansible_aws_profile,
            ansible_sym_resource=ansible_options.ansible_sym_resource,
            control_master=ansible_options.control_master,
            send_command=ansible_options.send_command,
            forks=ansible_options.forks,
        )



from .helpers.config import init as config_init

config_init("sym")

from .commands import import_all as command_init
from .commands.sym import sym  # noqa
from .helpers.init import sym_init

command_init()
sym_init()



from contextlib import contextmanager
from datetime import datetime, timedelta
from typing import List, Mapping, MutableMapping, Optional, TypedDict

import immutables

from sym.cli.helpers.config.base import ConfigBase, fail_ok


class ServerConfigSchema(TypedDict):
    region: Optional[str]
    last_connection: Optional[datetime]
    aliases: List[str]


class ConfigSchema(TypedDict, total=False):
    org: str
    email: str
    default_resource: str
    last_update_check: Optional[datetime]
    servers: MutableMapping[str, ServerConfigSchema]


class Config(ConfigBase[ConfigSchema]):
    @classmethod
    def get_org(cls) -> str:
        return cls.instance()["org"]

    @classmethod
    def get_handle(cls) -> str:
        return cls.get_email().split("@")[0]

    @classmethod
    def get_servers(cls) -> Mapping[str, ServerConfigSchema]:
        return cls.instance().get("servers", immutables.Map())

    @classmethod
    def get_instance(cls, instance: str) -> Optional[ServerConfigSchema]:
        return cls.get_servers().get(instance)

    @classmethod
    def find_instance_by_alias(
        cls, alias: str, ttl: timedelta = timedelta(days=1)
    ) -> Optional[str]:
        for instance, config in cls.get_servers().items():
            if alias in config.get("aliases", []) and cls.check_instance_ttl(
                instance,
                ttl,
                allow_new=True,
            ):
                return instance

    @classmethod
    def check_instance_ttl(
        cls,
        instance: str,
        ttl: timedelta = timedelta(days=1),
        allow_new=False,
    ) -> bool:
        instance_config = cls.get_instance(instance)
        if not instance_config:
            return allow_new
        last_connect = instance_config.get("last_connection")
        if not last_connect:
            return allow_new
        return datetime.now() - last_connect < ttl

    @classmethod
    @contextmanager
    def _atomic_instance_mutation(cls, instance: str):
        with cls.instance().atomic():
            instance_config = cls.get_instance(instance)
            if not instance_config:
                instance_config = ServerConfigSchema(
                    last_connection=None, region=None, aliases=[]
                )
            yield instance_config
            cls.instance()["servers"] = cls.get_servers().set(instance, instance_config)

    @classmethod
    @fail_ok
    def touch_instance(cls, instance: str, error: bool = False):
        with cls._atomic_instance_mutation(instance) as instance_config:
            if error:
                instance_config["last_connection"] = datetime.now() - timedelta(weeks=52)
                instance_config["aliases"] = []
            else:
                instance_config["last_connection"] = datetime.now()

    @classmethod
    @fail_ok
    def add_instance_alias(cls, instance: str, *, alias: str, region: str):
        if (old_instance := cls.find_instance_by_alias(alias)) :
            with cls._atomic_instance_mutation(old_instance) as instance_config:
                instance_config["aliases"].remove(alias)

        with cls._atomic_instance_mutation(instance) as instance_config:
            if "aliases" not in instance_config:
                instance_config["aliases"] = []
            if alias not in instance_config["aliases"]:
                instance_config["aliases"].append(alias)
            if region:
                instance_config["region"] = region

    @classmethod
    @fail_ok
    def record_instance(cls, instance_id: str, *, alias: str, region: str):
        with cls._atomic_instance_mutation(instance_id) as instance_config:
            if alias:
                instance_config["aliases"].append(alias)
            if region:
                instance_config["region"] = region

    @classmethod
    def is_sym(cls) -> bool:
        try:
            return cls.get_org() == "sym"
        except KeyError as e:
            return False

    @classmethod
    def check_for_update(cls, every=timedelta(days=1)):
        now = datetime.now()
        last_checked = cls.instance().get("last_update_check", now - every)
        if now - last_checked >= every:
            cls.instance()["last_update_check"] = now
            return True
        return False



from click.testing import Result
from expects.matchers import Matcher


class succeed(Matcher):
    def _match(self, result: Result):
        if result.exit_code == 0:
            return True, ["command succeeded"]
        else:
            errors = [f"command had exit code {result.exit_code}"]
            if result.output:
                errors.append(f"command failed: {result.output}")
            return False, errors


class fail_with(Matcher):
    def __init__(self, exception_klass):
        self.exception_klass = exception_klass

    def _match(self, result: Result):
        if result.exit_code == 0:
            return False, ["command succeeded"]
        elif result.exit_code != self.exception_klass.exit_code:
            return False, [f"command failed with exit code {result.exit_code}"]
        else:
            return True, [
                f"command failed with exit code {self.exception_klass.exit_code}"
            ]



import pytest
from expects import *

TEST_HOST = "test-host"
TEST_INSTANCE = "test-instance"


def describe_instances_stub(make_stub):
    ec2 = make_stub("ec2")
    ec2.add_response(
        "describe_instances",
        {"Reservations": [{"Instances": [{"InstanceId": TEST_INSTANCE}]}]},
    )


@pytest.fixture
def host_to_instance_tester(command_tester):
    return command_tester(["host-to-instance", "test", TEST_HOST])


def test_host_to_instance(
    host_to_instance_tester,
):
    def setup(make_stub):
        describe_instances_stub(make_stub)

    with host_to_instance_tester(setup=setup) as result:
        expect(result.output.strip()).to(equal(TEST_INSTANCE))



from dataclasses import dataclass
from typing import Optional, Tuple


@dataclass
class AnsibleOptions:
    command: Tuple[str, ...]
    control_master: bool
    send_command: bool
    forks: int
    ansible_aws_profile: Optional[str] = None
    ansible_sym_resource: Optional[str] = None



from sym.cli.helpers.os import has_command
from sym.cli.tests.helpers.sandbox import Sandbox


def test_has_command_not_found(sandbox: Sandbox) -> None:
    with sandbox.push_exec_path():
        assert not has_command("foo")


def test_has_command_found(sandbox: Sandbox) -> None:
    sandbox.create_file("bin/foo", 0o755)
    with sandbox.push_exec_path():
        assert has_command("foo")


def test_has_command_not_executable(sandbox: Sandbox) -> None:
    sandbox.create_file("bin/foo", 0o644)
    with sandbox.push_exec_path():
        assert not has_command("foo")



from sym.cli.saml_clients.saml_client import SAMLClient

from .cache import CachingEc2Client
from .client import Ec2Client


def get_ec2_client(client: SAMLClient) -> Ec2Client:
    return CachingEc2Client(client)



# Ignore black formatting in this file to ensure backward
# compatibility with python 2.7
# fmt: off

from __future__ import absolute_import, division, print_function

__metaclass__ = type

DOCUMENTATION = """
author:
- Yasyf Mohamedali <yasyf@symops.io>
- Rick Ducott <rick@symops.io>
connection: sym_aws_ssm
short_description: Use SSM SendCommand as an Ansible Connection
description:
- Use SSM SendCommand as an Ansible Connection. Inspired heavily by community.aws.aws_ssm.
requirements:
- session-manager-plugin is on your PATH
options:
  region:
    description: The region to use for all boto commands
    vars:
    - name: ansible_aws_ssm_region
    default: 'us-east-1'
  bucket:
    description: The name of the S3 bucket used for file transfers.
    vars:
    - name: ansible_aws_ssm_bucket
  profile:
    description: The AWS profile to use
    vars:
    - name: ansible_aws_ssm_profile
      default: 'sym-ansible'
  retries:
    description: Number of attempts to connect.
    default: 3
    type: integer
    vars:
    - name: ansible_aws_ssm_retries
  ssm_timeout:
    description: SSM connection timeout (in seconds)
    default: 60
    type: integer
    vars:
    - name: ansible_aws_ssm_timeout
  host_command:
    description: Host-to-Instance command
    vars:
    - name: ansible_aws_ssm_host_cmd
"""

import json
import os
import pty
import select
import shlex
import subprocess
import time
from functools import wraps

import boto3
from ansible.errors import AnsibleConnectionFailure, AnsibleError, AnsibleFileNotFound
from ansible.module_utils._text import to_bytes, to_native, to_text
from ansible.module_utils.six.moves import xrange
from ansible.plugins.connection import ConnectionBase
from ansible.utils.display import Display

URL_TTL = 3600
CHUNK_SIZE = 1024
MAX_DELAY = 30
START_COMMAND = "SYM_COMMAND_START"
END_COMMAND = "SYM_COMMAND_END"

display = Display()


def ssm_retry(func):
    @wraps(func)
    def wrapped(self, *args, **kwargs):
        retries = int(self.get_option("retries"))
        cmd = args[0]

        for attempt in xrange(retries):
            try:
                response = func(self, *args, **kwargs)
                display.vvv(response, host=self.host)
                return response
            except self._ssm_client.exceptions.TargetNotConnected as e:
                raise AnsibleConnectionFailure(
                    "Instance %s (%s) must be managed by Session Manager. Is SSM Agent installed and active?"
                    % (self.instance_id, self.host),
                    orig_exc=e,
                )
            except (AnsibleConnectionFailure, Exception) as e:
                if "RequestExpired" in str(e):
                    raise AnsibleConnectionFailure(
                        "Your AWS credentials have expired. Please refresh them and try again.",
                        orig_exc=e,
                    )

                if attempt == retries - 1:
                    raise AnsibleConnectionFailure(e)

                delay = min((2 ** attempt) - 1, MAX_DELAY)

                if isinstance(e, AnsibleConnectionFailure):
                    msg = "ssm_retry: attempt: %d, cmd (%s), pausing for %d seconds" % (
                        attempt,
                        cmd,
                        delay,
                    )
                else:
                    msg = (
                        "ssm_retry: attempt: %d, caught exception(%s) from cmd (%s), pausing for %d seconds"
                        % (attempt, e, cmd, delay)
                    )
                display.vv(msg, host=self.host)

                time.sleep(delay)
                self.close()

    return wrapped


def chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


class Poller:
    def __init__(self, fd):
        self.fd = fd

    @classmethod
    def create(cls, fd):
        if hasattr(select, "poll"):
            return PollPoller(fd)
        else:
            return KQueuePoller(fd)

    def poll(self, timeout):
        return self._poll(timeout=timeout)

    def _poll(self, timeout):
        raise NotImplementedError


class KQueuePoller(Poller):
    def __init__(self, fd):
        super(KQueuePoller, self).__init__(fd)
        self.kq = select.kqueue()
        self.kevts = [select.kevent(self.fd)]

    def _poll(self, timeout):
        return bool(self.kq.control(self.kevts, 1, timeout))


class PollPoller(Poller):
    def __init__(self, fd):
        super(PollPoller, self).__init__(fd)
        self.poller = select.poll()
        self.poller.register(self.fd, select.POLLIN)

    def _poll(self, timeout):
        return bool(self.poller.poll(timeout))


class Connection(ConnectionBase):
    """ AWS SSM based connections """

    transport = "sym_aws_ssm"
    allow_executable = False
    allow_extras = True
    has_pipelining = False

    _process = None
    _session_id = None

    _stdout = None
    _poll_stdout = None
    _poll_stderr = None

    _ssm_client = None
    _s3_client = None

    _terminate = False

    def __init__(self, *args, **kwargs):
        super(Connection, self).__init__(*args, **kwargs)
        self.host = self._play_context.remote_addr
        self.user = self._play_context.remote_user

        # In Ansible 2.5, _load_name has not yet been set,
        # so the loading resources that use `self.get_option` from
        # init fail. As a result, we set them to None here and
        # initialize them when they're needed.
        self._ssm_client = None
        self._s3_client = None

    def _connect(self):
        if not self._session_id:
            self.start_ssm_session()
        return self

    def _get_instance_meta(self):
        cmd = self.get_option("host_command").replace("$HOST", self.host)
        return json.loads(to_text(subprocess.check_output(shlex.split(cmd)).strip()))

    def _start_process(self, cmd):
        display.vvvv(u"SSM COMMAND: {0}".format(to_text(cmd)), host=self.host)

        stdout_r, stdout_w = pty.openpty()
        self._process = subprocess.Popen(
            cmd,
            stdin=subprocess.PIPE,
            stdout=stdout_w,
            stderr=subprocess.PIPE,
            close_fds=True,
            bufsize=0,
        )
        os.close(stdout_w)

        self._stdout = os.fdopen(stdout_r, "rb", 0)

        self._poll_stdout = Poller.create(self._stdout)
        self._poll_stderr = Poller.create(self._process.stderr)

    def _prepare_terminal(self):
        # Disable command echo and prompt
        self._process.stdin.write(to_bytes("stty -echo\n" + "PS1=''\n"))

    def start_ssm_session(self):
        instance_meta = self._get_instance_meta()
        self.instance_id = instance_meta["instance"]

        self._s3_client = self._get_boto_client("s3")
        self._ssm_client = self._get_boto_client("ssm", region=instance_meta["region"])

        display.vvv(
            u"ESTABLISH SSM CONNECTION TO: {0}".format(self.instance_id),
            host=self.host,
        )

        response = self._ssm_client.start_session(
            Target=self.instance_id,
            DocumentName="AWS-StartInteractiveCommand",
            Parameters={"command": ["sudo su -l {0}".format(self.user)]},
        )
        self._session_id = response["SessionId"]
        display.vvv(u"SSM CONNECTION ID: {0}".format(self._session_id), host=self.host)

        cmd = [
            "session-manager-plugin",
            json.dumps(response),
            self.get_option("region"),
            "StartSession",
            self.get_option("profile"),
            json.dumps({"Target": self.instance_id}),
            self._ssm_client.meta.endpoint_url,
        ]

        self._start_process(cmd)
        self._prepare_terminal()

    def _wrap_command(self, cmd):
        return "echo {1}; {0}; echo {2} $?\n".format(cmd, START_COMMAND, END_COMMAND)

    def _flush_stderr(self):
        stderr = ""

        while self._process.poll() is None:
            if self._poll_stderr.poll(1):
                line = self._process.stderr.readline()
                display.vvvv(u"stderr line: {0}".format(to_text(line)), host=self.host)
                stderr = stderr + line
            else:
                break

        return stderr

    def _get_returncode(self, line):
        return int(line.split(" ")[-1])

    def _process_line(self, line):
        return to_text(line).replace("\r\r\n", "\n")

    @ssm_retry
    def exec_command(self, cmd, in_data=None, sudoable=True):
        super(Connection, self).exec_command(cmd, in_data=in_data, sudoable=sudoable)

        cmd = self._wrap_command(cmd)
        display.vvv(u"EXEC send: {0}".format(to_text(cmd)), host=self.host)

        self._flush_stderr()

        for chunk in chunks(cmd, CHUNK_SIZE):
            self._process.stdin.write(to_bytes(chunk))

        started = False
        stdout = ""

        for elapsed in range(self.get_option("ssm_timeout")):
            if self._process.poll():
                break

            if elapsed == self.get_option("ssm_timeout") - 1:
                self._terminate = True
                display.vvvv(
                    u"EXEC timeout stdout: {0}".format(to_text(stdout)),
                    host=self.host,
                )
                raise AnsibleConnectionFailure(
                    "SSM exec_command timeout on host: {0}".format(self.instance_id)
                )

            if not self._poll_stdout.poll(1000):
                display.vvvv(u"EXEC elapsed: {0}".format(elapsed), host=self.host)
                continue

            line = self._process_line(self._stdout.readline())
            display.vvvv(u"EXEC stdout line: {0}".format(line), host=self.host)

            if not started and START_COMMAND in line:
                display.vvvv(u"EXEC start", host=self.host)
                started = True
                continue

            if not started:
                continue

            if END_COMMAND in line:
                returncode = self._get_returncode(line)
                display.vvvv(u"EXEC end: {0}".format(line), host=self.host)
                return (returncode, stdout, self._flush_stderr())

            stdout = stdout + line

        return (self._process.returncode, stdout, self._flush_stderr())

    def _get_boto_client(self, service, region=None):
        return boto3.Session(
            profile_name=self.get_option("profile"),
            region_name=region or self.get_option("region"),
        ).client(service)

    def _get_url(self, action, key, http_method):
        return self._s3_client.generate_presigned_url(
            action,
            Params={"Bucket": self.get_option("bucket"), "Key": key},
            ExpiresIn=URL_TTL,
            HttpMethod=http_method,
        )

    @ssm_retry
    def _file_transport_command(self, in_path, out_path, ssm_action):
        path = "{0}/{1}".format(self.instance_id, out_path)

        if ssm_action == "get":
            command = "curl --request PUT --upload-file '{0}' '{1}'".format(
                in_path, self._get_url("put_object", path, "PUT")
            )
            result = self.exec_command(command, in_data=None, sudoable=False)
            with open(to_bytes(out_path), "wb") as f:
                self._s3_client.download_fileobj(self.get_option("bucket"), path, f)
        else:
            command = "curl '{0}' -o '{1}'".format(
                self._get_url("get_object", path, "GET"), out_path
            )
            with open(to_bytes(in_path), "rb") as f:
                self._s3_client.upload_fileobj(f, self.get_option("bucket"), path)
            result = self.exec_command(command, in_data=None, sudoable=False)

        self._s3_client.delete_object(Bucket=self.get_option("bucket"), Key=path)

        (returncode, stdout, stderr) = result
        if returncode == 0:
            return (returncode, stdout, stderr)
        else:
            raise AnsibleError(
                "failed to transfer file to %s %s:\n%s\n%s"
                % (
                    to_native(in_path),
                    to_native(out_path),
                    to_native(stdout),
                    to_native(stderr),
                )
            )

    def put_file(self, in_path, out_path):
        super(Connection, self).put_file(in_path, out_path)
        display.vvv(u"PUT {0} TO {1}".format(in_path, out_path), host=self.host)
        if not os.path.exists(to_bytes(in_path)):
            raise AnsibleFileNotFound(
                "file or module does not exist: {0}".format(to_native(in_path))
            )
        return self._file_transport_command(in_path, out_path, "put")

    def fetch_file(self, in_path, out_path):
        super(Connection, self).fetch_file(in_path, out_path)
        display.vvv(u"FETCH {0} TO {1}".format(in_path, out_path), host=self.host)
        return self._file_transport_command(in_path, out_path, "get")

    def close(self):
        if not self._session_id:
            return

        display.vvv(u"CLOSE SSM: {0}".format(self.instance_id), host=self.host)

        if self._terminate:
            self._process.terminate()
        else:
            try:
                self._process.communicate(b"\nexit\n")
            except Exception as e:  # Not sure why this fails sometimes
                display.vvv(u"ERROR CLOSING SSM: {0}".format(e), host=self.host)
                self._process.terminate()

        display.vvvv(u"TERMINATE SSM: {0}".format(self._session_id), host=self.host)
        self._ssm_client.terminate_session(SessionId=self._session_id)
        self._session_id = None



import os
from configparser import ConfigParser
from functools import cached_property
from pathlib import Path
from typing import Iterator, Tuple

from ..decorators import command_require_bins, intercept_errors, run_subprocess
from ..helpers.boto import get_identity
from ..helpers.contexts import push_env
from ..helpers.keywords_to_options import Argument
from ..helpers.params import Profile
from .saml_client import SAMLClient

AwsCredentialsPath = Path(
    os.getenv("AWS_CREDENTIAL_FILE", Path.home() / ".aws" / "credentials")
)


class AwsProfile(SAMLClient):
    binary = "aws"
    option_value = "aws-profile"
    priority = 0
    setup_help = f"Set up your profile in `{str(AwsCredentialsPath)}`."

    resource: str
    options: "GlobalOptions"

    @classmethod
    def validate_resource(cls, resource: str):
        config = ConfigParser()
        config.read(AwsCredentialsPath)
        return config.has_section(resource)

    @cached_property
    def _section(self):
        config = ConfigParser()
        config.read(AwsCredentialsPath)
        return config[self.resource]

    def get_creds(self):
        creds = {k.upper(): v for k, v in self._section.items() if k.startswith("aws")}
        creds["AWS_REGION"] = self._section.get("region")
        creds["AWS_CREDENTIAL_EXPIRATION"] = self._section.get("x_security_token_expires")
        return creds

    @intercept_errors()
    @run_subprocess
    @command_require_bins(binary)
    def _exec(self, *args: str, **opts: str) -> Iterator[Tuple[Argument, ...]]:
        self.log_subprocess_event(args)
        with push_env("AWS_PROFILE", self.resource):
            yield (*args, opts)

    def is_setup(self) -> bool:
        return AwsCredentialsPath.exists()

    def _ensure_config(self, profile: Profile) -> ConfigParser:
        return ConfigParser()

    def _ensure_session(self, *, force: bool):
        if not force and not self._creds_expiring():
            return
        get_identity(self)

    def get_profile(self):
        return Profile(
            display_name=self.resource,
            region=self._section.get("region"),
            arn=self._section.get("x_principal_arn"),
            ansible_bucket=self._section.get("x_sym_ansible_bucket"),
        )



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Optional, Sequence

from sym.cli.data.global_options_base import GlobalOptionsBase
from sym.cli.errors import CliError
from sym.cli.helpers.ec2.client import RunningInstance


@dataclass
class CheckContext:
    options: GlobalOptionsBase
    resource: str
    regions: Sequence[str] = field(default_factory=list)
    running_instances: Sequence[RunningInstance] = field(default_factory=list)


class CheckError(CliError):
    def __init__(self) -> None:
        super().__init__(f"Checks failed!")


@dataclass
class CheckResult:
    error: Optional[Exception] = None
    success: bool = False
    msg: str = ""


class SymCheck(ABC):
    @abstractmethod
    def check(self, ctx: CheckContext) -> CheckResult:
        raise NotImplementedError()


def success(msg: str) -> CheckResult:
    result = CheckResult()
    result.success = True
    result.msg = msg
    return result


def failure(msg: str) -> CheckResult:
    result = CheckResult()
    result.msg = msg
    return result



import pytest

from sym.cli.errors import SuppressedError
from sym.cli.helpers.config import Config
from sym.cli.tests.helpers.capture import CaptureCommand
from sym.cli.tests.helpers.ec2.conftest import TEST_REGION

INSTANCE_ID = "123"
SESSION_ID = "456"


def update_config():
    Config.record_instance(INSTANCE_ID, region=TEST_REGION, alias="10.20.30.40")


def start_session_stub(make_stub):
    ssm = make_stub("ssm")
    ssm.add_response("start_session", {"SessionId": SESSION_ID})
    ssm.add_response(
        "terminate_session", {"SessionId": SESSION_ID}, {"SessionId": SESSION_ID}
    )


@pytest.fixture
def ssh_session_tester(command_tester):
    return command_tester(["ssh-session", "test", "--instance", INSTANCE_ID])


def test_ssh_session(ssh_session_tester, capture_command: CaptureCommand):
    def setup(make_stub):
        update_config()
        start_session_stub(make_stub)

    with ssh_session_tester(setup=setup):
        capture_command.assert_command(
            ["exec", "true"],
            ["exec", "true"],
            ["session-manager-plugin"],
        )


def test_ssh_session_error(ssh_session_tester, capture_command: CaptureCommand):
    def setup(make_stub):
        update_config()
        start_session_stub(make_stub)
        capture_command.register_output(r"session-manager-plugin", "timeout", exit_code=1)

    with ssh_session_tester(setup=setup, exception=SuppressedError):
        capture_command.assert_command(
            ["exec", "true"],
            ["exec", "true"],
            ["session-manager-plugin"],
        )



from contextlib import contextmanager
from os import fchmod, get_exec_path, symlink
from pathlib import Path
from typing import Iterator, Optional, TextIO

from sym.cli.helpers.config import Config
from sym.cli.helpers.contexts import push_env
from sym.cli.helpers.os import find_command


class Sandbox:
    def __init__(self, path: Path) -> None:
        self.path = path
        self.exec_path = get_exec_path()

    def link_binary(self, name):
        symlink(find_command(name, path=self.exec_path), str(self.path / "bin" / name))

    def create_binary(self, name, exist_ok=True):
        try:
            self.create_file(name, mode=0o755, contents="#!/bin/bash\n")
        except FileExistsError:
            if not exist_ok:
                raise

    def create_dir(self, name: str) -> Path:
        path = self.path / name
        path.mkdir(parents=True, exist_ok=True)
        return path

    def create_file(
        self, name: str, mode: int = 0o644, contents: Optional[str] = None
    ) -> None:
        with self.create_file_with_content(name, mode) as f:
            if contents:
                f.write(contents)

    @contextmanager
    def create_file_with_content(self, name: str, mode: int = 0o644) -> Iterator[TextIO]:
        path = self.path / name
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open("x") as f:
            fchmod(f.fileno(), mode)
            yield f

    @contextmanager
    def push_exec_path(self) -> Iterator[None]:
        with push_env("PATH", str(self.path / "bin")):
            yield

    @contextmanager
    def push_xdg_config_home(self) -> Iterator[None]:
        with push_env("XDG_CONFIG_HOME", str(self.path / ".config")):
            Config.reset()
            yield

    @contextmanager
    def push_home(self) -> Iterator[None]:
        with push_env("HOME", str(self.create_dir("home"))):
            yield



from os import environ
from typing import Callable, Iterator
from uuid import UUID

import pytest

from sym.cli.helpers.contexts import push_env


@pytest.fixture
def fresh_env_var(uuid_factory: Callable[[], UUID]) -> Iterator[str]:
    uuid = uuid_factory().hex
    # Extremely unlikely, but just in case:
    while uuid in environ:
        uuid = uuid_factory().hex
    try:
        yield uuid
    finally:
        if uuid in environ:
            del environ[uuid]


def _assert_environ(name: str, value: str) -> None:
    assert environ[name] == value


def _assert_environ_absent(name: str) -> None:
    assert name not in environ


def test_simple(fresh_env_var: str) -> None:
    environ[fresh_env_var] = "foo"
    with push_env(fresh_env_var, "bar"):
        _assert_environ(fresh_env_var, "bar")
    _assert_environ(fresh_env_var, "foo")


def test_from_nil(fresh_env_var: str) -> None:
    with push_env(fresh_env_var, "foo"):
        _assert_environ(fresh_env_var, "foo")
    _assert_environ_absent(fresh_env_var)


def test_to_nil(fresh_env_var: str) -> None:
    environ[fresh_env_var] = "foo"
    with push_env(fresh_env_var, None):
        _assert_environ_absent(fresh_env_var)
    _assert_environ(fresh_env_var, "foo")


def test_from_to_nil(fresh_env_var: str) -> None:
    with push_env(fresh_env_var, None):
        _assert_environ_absent(fresh_env_var)
    _assert_environ_absent(fresh_env_var)



from typing import List

from sym.cli.data.global_options_base import GlobalOptionsBase
from sym.cli.helpers.check.model import CheckContext, CheckError, SymCheck
from sym.cli.helpers.check.ssh import (
    EnsureSshKeyAuthorizedCheck,
    GenSshConfigCheck,
    GenSshKeyCheck,
)

from .core import DependenciesCheck, LoginCheck, ResourceCheck
from .ec2 import CallerIdentityCheck, DescribeInstanceCheck, DescribeRegionsCheck


def run_check(check: SymCheck, ctx: CheckContext) -> bool:
    result = check.check(ctx)
    if result.success:
        print("\u2713", result.msg)
        return True
    else:
        print("\u2717", result.msg)
        if result.error:
            print(result.error)
        return False


def get_checks(instances: List[str]) -> List[SymCheck]:
    checks = [
        LoginCheck(),
        DependenciesCheck("aws"),
        DependenciesCheck("session-manager-plugin"),
        DependenciesCheck("ssh-keygen"),
        ResourceCheck(),
        CallerIdentityCheck(),
        DescribeRegionsCheck(),
        GenSshConfigCheck(),
        GenSshKeyCheck(),
    ]
    for instance in instances:
        checks.append(DescribeInstanceCheck(instance))
        checks.append(EnsureSshKeyAuthorizedCheck(instance))
    return checks


def must_run_all(options: GlobalOptionsBase, resource: str, instances: List[str]):
    """
    Runs all checks with the supplied args.
    Raises a CheckError on the first check failure, or returns normally if all checks succeed.
    """
    ctx = CheckContext(options=options, resource=resource)
    checks = get_checks(instances)
    for check in checks:
        if not run_check(check, ctx):
            raise CheckError()
